import pickle
from typing import Any, Dict, List

import depthai as dai
import numpy as np

from depthai_nodes import (
    Classifications,
    Clusters,
    ImgDetectionsExtended,
    Keypoints,
    Lines,
    Map2D,
    Predictions,
    SegmentationMask,
)

from .utils import extract_main_slug


def load_expected_output(model: str, parser: str) -> Dict[str, Any]:
    model = extract_main_slug(model)
    with open(f"nn_datas/{parser}/{model}_output.pkl", "rb") as f:
        return pickle.load(f)


def check_classification_msg(
    message: Classifications, expected_output: Dict[str, Any], verbose: bool = False
):
    """
    Expected output format:
    {
        "model": "luxonis/efficientnet-lite:lite0-224x224",
        "parser": "ClassificationParser",
        "class": "car",
        "score": 0.9
    }
    """
    assert isinstance(
        message, Classifications
    ), f"The message is not a Classifications. Got {type(message)}."

    if verbose:
        print(
            f"Expected top class: {expected_output['class']}, predicted top class: {message.top_class}"
        )
        print(
            f"Expected top score: {expected_output['score']}, predicted top score: {message.top_score}"
        )
    assert message.top_class == expected_output["class"]
    np.testing.assert_allclose(message.top_score, expected_output["score"], rtol=1e-2)


def check_classification_sequence_msg(
    message: Classifications, expected_output: Dict[str, Any], verbose: bool = False
):
    """
    Expected output format:
    {
        "model": "luxonis/efficientnet-lite:lite0-224x224",
        "parser": "ClassificationSequenceParser",
        "class": ['HELLO']
    """
    assert isinstance(
        message, Classifications
    ), f"The message is not a Classifications. Got {type(message)}."

    if verbose:
        print(
            f"Expected top class: {expected_output['class']}, predicted top class: {message.classes}"
        )
    assert len(message.classes) == len(
        expected_output["class"]
    ), "The number of classes is different."
    for i in range(len(message.classes)):
        assert (
            message.classes[i] == expected_output["class"][i]
        ), f"Class {i} is different."


def check_embeddings_msg(
    message: dai.NNData, expected_output: Dict[str, Any], verbose: bool = False
):
    """
    Expected output format:
    {
        "model": "luxonis/arcface:lfw-112x112",
        "parser": "EmbeddingsParser",
        "embeddings": array([[-9.47265625e-02, -1.23901367e-01,...]])
    }
    """
    assert isinstance(
        message, dai.NNData
    ), f"The message is not a dai.NNData. Got {type(message)}."

    outputs = message.getAllLayerNames()
    assert len(outputs) == 1, "The number of outputs must be 1."
    embeddings = message.getTensor(outputs[0])
    expected_embeddings = expected_output["embeddings"]
    if verbose:
        print(
            f"Expected embeddings shape: {expected_embeddings.shape}, predicted embeddings shape: {embeddings.shape}"
        )
    assert (
        embeddings.shape == expected_embeddings.shape
    ), "The shape of the embeddings is different."
    np.testing.assert_allclose(embeddings, expected_embeddings, rtol=1e-2)


def check_segmentation_msg(
    message: SegmentationMask,
    expected_output: Dict[str, Any],
    threshold: float = 0.9,
    verbose: bool = False,
):
    """
    Expected output format:
    {
        "model": "luxonis/fastsam-s:512x288",
        "parser": "FastSAMParser",
        "mask": np.array([[0, 0, 0, ..., 0, 0, 0]])
    }
    """
    assert isinstance(
        message, SegmentationMask
    ), f"The message is not a SegmentationMask. Got {type(message)}."

    mask = message.mask

    expected_mask = expected_output["mask"]
    if verbose:
        print(
            f"Expected mask shape: {expected_mask.shape}, predicted mask shape: {mask.shape}"
        )
    assert (
        mask.shape == expected_mask.shape
    ), f"The shape of the mask is different. Expects {expected_mask.shape}, got {mask.shape}"
    assert (
        mask.dtype == expected_mask.dtype
    ), f"The dtype of the mask is different. Expects {expected_mask.dtype}, got {mask.dtype}"

    correct = (mask == expected_mask).sum()
    total = expected_mask.size
    acc = correct / total
    if verbose:
        print(f"Accuracy: {acc}")
    assert (
        acc > threshold
    ), f"The accuracy {acc} is lower than the threshold {threshold}"


def check_keypoints_msg(
    message: Keypoints,
    expected_output: Dict[str, Any],
    verbose: bool = False,
):
    """
    Expected output format:
    {
        "model": "luxonis/mediapipe-face-landmarker:192x192",
        "parser": "KeypointParser",
        "keypoints": [[0.1, 0.2], ...]
    """
    assert isinstance(
        message, Keypoints
    ), f"The message is not a Keypoints. Got {type(message)}."

    keypoints = [[kp.x, kp.y] for kp in message.keypoints]
    keypoints = np.array(keypoints)
    expected_keypoints = np.array(expected_output["keypoints"])
    if expected_keypoints.shape[1] == 3:
        expected_keypoints = expected_keypoints[:, :2]  # we dont need the confidences
    if verbose:
        print(
            f"Expected keypoints shape: {expected_keypoints.shape}, predicted keypoints shape: {keypoints.shape}"
        )
    assert (
        keypoints.shape == expected_keypoints.shape
    ), f"The shape of the keypoints is different. Expects {expected_keypoints.shape}, got {keypoints.shape}"

    np.testing.assert_allclose(keypoints, expected_keypoints, rtol=1e-2)


def check_image_msg(
    message: dai.ImgFrame, expected_output: Dict[str, Any], verbose: bool = False
):
    """
    Expected output format:
    {
        "model": "luxonis/dncnn3:240x320",
        "parser": "ImageOutputParser",
        "output": np.array([[0, 0, 0, ..., 0, 0, 0]])
    }
    """
    assert isinstance(
        message, dai.ImgFrame
    ), f"The message is not a dai.ImgFrame. Got {type(message)}."

    image = message.getCvFrame()
    expected_image = expected_output["output"]
    if verbose:
        print(
            f"Expected image shape: {expected_image.shape}, predicted image shape: {image.shape}"
        )
    assert (
        image.shape == expected_image.shape
    ), f"The shape of the image is different. Expects {expected_image.shape}, got {image.shape}"
    assert np.allclose(image, expected_image, atol=1), "The image is different."


def check_cluster_msg(
    message: Clusters, expected_output: Dict[str, Any], verbose: bool = False
):
    """
    Expected output format:
    {
        "model": "luxonis/ultra-fast-lane-detection:culane-800x288",
        "parser": "LaneDetectionParser",
        "clusters": [[[0.1, 0.2], ...]]
    """
    assert isinstance(
        message, Clusters
    ), f"The message is not a Clusters. Got {type(message)}."

    clusters = message.clusters
    expected_clusters = expected_output["clusters"]
    clusters_list = []
    for cluster in clusters:
        c = []
        for point in cluster.points:
            c.append([point.x, point.y])
        clusters_list.append(c)

    # check if all points are in the expected clusters
    for i, cluster in enumerate(expected_clusters):
        for gt_point in cluster:
            found = False
            for pred_point in clusters_list[i]:
                if np.allclose(gt_point, pred_point, atol=0.001):
                    found = True
                    break
            assert found, f"Expected point {gt_point} not found in the cluster."


def check_map_msg(
    message: Map2D, expected_output: Dict[str, Any], verbose: bool = False
):
    """
    Expected output format:
    {
        "model": "luxonis/dm-count:sha-144x256",
        "parser": "MapOutputParser",
        "map": np.array([[0, 0, 0, ..., 0, 0, 0]])
    }
    """
    assert isinstance(
        message, Map2D
    ), f"The message is not a Map2D. Got {type(message)}."

    map_tensor = message.map
    expected_map = expected_output["map"]
    if verbose:
        print(
            f"Expected map shape: {expected_map.shape}, predicted map shape: {map_tensor.shape}"
        )
    assert (
        map_tensor.shape == expected_map.shape
    ), f"The shape of the map is different. Expects {expected_map.shape}, got {map_tensor.shape}"
    assert np.allclose(map_tensor, expected_map, atol=0.001), "The map is different."


def check_detection_msg(
    message: ImgDetectionsExtended,
    expected_output: Dict[str, Any],
    verbose: bool = False,
):
    """
    Expected output format:
    {
        "model": "luxonis/scrfd-face-detection:10g-640x640",
        "parser": "SCRFDParser",
        "detections": [
            {
                "confidence": 0.9,
                "label": "person",
                "x_center": 0.1,
                "y_center": 0.2,
                "width": 0.3,
                "height": 0.4,
                "angle": 0.5,
                "keypoints": [[0.1, 0.2, 0.3], ...],
                "mask": np.array([[0, 0, 0, ..., 0, 0, 0]])
            },
            ...
        ]
    }
    """
    assert isinstance(
        message, ImgDetectionsExtended
    ), f"The message is not a ImgDetectionsExtended. Got {type(message)}."

    predicted_detections = []
    expected_detections: List[Dict[str, Any]] = expected_output["detections"]

    for detection in message.detections:
        d = {
            "confidence": detection.confidence,
            "label": detection.label,
            "x_center": detection.rotated_rect.center.x,
            "y_center": detection.rotated_rect.center.y,
            "width": detection.rotated_rect.size.width,
            "height": detection.rotated_rect.size.height,
            "angle": detection.rotated_rect.angle,
            "keypoints": [[kp.x, kp.y, kp.confidence] for kp in detection.keypoints],
            "mask": message.masks,
        }
        predicted_detections.append(d)

    assert (
        len(predicted_detections) == len(expected_detections)
    ), f"The number of detections is different. Got {len(predicted_detections)}, expected {len(expected_detections)}"

    if verbose:
        print(
            f"Expected number of detections: {len(expected_detections)}, predicted number of detections: {len(predicted_detections)}"
        )

    for i, expected_detection in enumerate(expected_detections):
        predicted_detection = predicted_detections[i]
        assert np.allclose(
            expected_detection["confidence"],
            predicted_detection["confidence"],
            atol=0.01,
        ), f"Expected confidence {expected_detection['confidence']}, got {predicted_detection['confidence']}."
        assert (
            expected_detection["label"] == predicted_detection["label"]
        ), f"Expected label {expected_detection['label']}, got {predicted_detection['label']}."
        assert np.allclose(
            expected_detection["x_center"], predicted_detection["x_center"], atol=0.01
        ), f"Expected x_center {expected_detection['x_center']}, got {predicted_detection['x_center']}."
        assert np.allclose(
            expected_detection["y_center"], predicted_detection["y_center"], atol=0.01
        ), f"Expected y_center {expected_detection['y_center']}, got {predicted_detection['y_center']}."
        assert np.allclose(
            expected_detection["width"], predicted_detection["width"], atol=0.01
        ), f"Expected width {expected_detection['width']}, got {predicted_detection['width']}."
        assert np.allclose(
            expected_detection["height"], predicted_detection["height"], atol=0.01
        ), f"Expected height {expected_detection['height']}, got {predicted_detection['height']}."
        assert np.allclose(
            expected_detection["angle"], predicted_detection["angle"], atol=0.01
        ), f"Expected angle {expected_detection['angle']}, got {predicted_detection['angle']}."

        if "keypoints" in expected_detection.keys():
            assert np.allclose(
                expected_detection["keypoints"],
                predicted_detection["keypoints"],
                atol=0.01,
            ), f"Expected keypoints {expected_detection['keypoints']}, got {predicted_detection['keypoints']}."

        if "mask" in expected_detection.keys():
            if verbose:
                print(
                    f"Expected mask shape: {expected_detection['mask'].shape}, predicted mask shape: {predicted_detection['mask'].shape}"
                )
            assert (
                expected_detection["mask"].shape == predicted_detection["mask"].shape
            ), f"The shape of the mask is different. Expects {expected_detection['mask'].shape}, got {predicted_detection['mask'].shape}"
            assert np.allclose(
                expected_detection["mask"], predicted_detection["mask"], atol=0.01
            ), f"Expected mask {expected_detection['mask']}, got {predicted_detection['mask']}."


def check_line_msg(
    message: Lines, expected_output: Dict[str, Any], verbose: bool = False
):
    """
    Expected output format:
    {
        "model": "luxonis/m-mlsd-tiny:512x512",
        "parser": "MLSDParser",
        "lines": [
            {
                "confidence": 0.9,
                "start_point": [0.1, 0.2],
                "end_point": [0.3, 0.4]
            },
            ...
        ]
    }
    """
    assert isinstance(
        message, Lines
    ), f"The message is not a Lines. Got {type(message)}."

    expected_lines: List[Dict[str, Any]] = expected_output["lines"]
    predicted_lines = []
    for line in message.lines:
        line_dict = {
            "confidence": line.confidence,
            "start_point": [line.start_point.x, line.start_point.y],
            "end_point": [line.end_point.x, line.end_point.y],
        }
        predicted_lines.append(line_dict)

    if verbose:
        print(
            f"Expected number of lines: {len(expected_lines)}, predicted number of lines: {len(predicted_lines)}"
        )
    assert (
        len(predicted_lines) == len(expected_lines)
    ), f"The number of lines is different. Got {len(predicted_lines)}, expected {len(expected_lines)}"

    for i, expected_line in enumerate(expected_lines):
        predicted_line = predicted_lines[i]
        assert np.allclose(
            expected_line["confidence"], predicted_line["confidence"], atol=0.01
        ), f"Expected confidence {expected_line['confidence']}, got {predicted_line['confidence']}."
        assert np.allclose(
            expected_line["start_point"], predicted_line["start_point"], atol=0.01
        ), f"Expected start_point {expected_line['start_point']}, got {predicted_line['start_point']}."
        assert np.allclose(
            expected_line["end_point"], predicted_line["end_point"], atol=0.01
        ), f"Expected end_point {expected_line['end_point']}, got {predicted_line['end_point']}."


def check_regression_msg(
    message: Predictions, expected_output: Dict[str, Any], verbose: bool = False
):
    """
    Expected output format:
    {
        "model": "luxonis/gaze-estimation-adas:60x60:0.0.1",
        "parser": "RegressionParser",
        "value": [0.4, 0.2, 0.1]
    }
    """
    assert isinstance(
        message, Predictions
    ), f"The message is not a Predictions. Got {type(message)}."

    predictions = message.predictions
    predictions = np.array([pred.prediction for pred in predictions])
    expected_predictions = np.array(expected_output["value"])

    if verbose:
        print(
            f"Expected predictions shape: {expected_predictions.shape}, predicted predictions shape: {predictions.shape}"
        )
    assert (
        predictions.shape == expected_predictions.shape
    ), f"The shape of the predictions is different. Expects {expected_predictions.shape}, got {predictions.shape}"
    np.testing.assert_allclose(predictions, expected_predictions, rtol=1e-2)


def check_output(message, model_slug, parser_name):
    expected_output = load_expected_output(model_slug, parser_name)

    if expected_output["parser"] == "ClassificationParser":
        check_classification_msg(message, expected_output)
    elif expected_output["parser"] == "ClassificationSequenceParser":
        check_classification_sequence_msg(message, expected_output)
    elif expected_output["parser"] == "EmbeddingsParser":
        check_embeddings_msg(message, expected_output)
    elif expected_output["parser"] == "FastSAMParser":
        check_segmentation_msg(message, expected_output)
    elif expected_output["parser"] == "HRNetParser":
        check_keypoints_msg(message, expected_output)
    elif expected_output["parser"] == "KeypointParser":
        check_keypoints_msg(message, expected_output)
    elif expected_output["parser"] == "ImageOutputParser":
        check_image_msg(message, expected_output)
    elif expected_output["parser"] == "LaneDetectionParser":
        check_cluster_msg(message, expected_output)
    elif expected_output["parser"] == "MapOutputParser":
        check_map_msg(message, expected_output)
    elif expected_output["parser"] == "MPPalmDetectionParser":
        check_detection_msg(message, expected_output)
    elif expected_output["parser"] == "MLSDParser":
        check_line_msg(message, expected_output)
    elif expected_output["parser"] == "PPTextDetectionParser":
        check_detection_msg(message, expected_output)
    elif expected_output["parser"] == "RegressionParser":
        check_regression_msg(message, expected_output)
    elif expected_output["parser"] == "SCRFDParser":
        check_detection_msg(message, expected_output)
    elif expected_output["parser"] == "SegmentationParser":
        check_segmentation_msg(message, expected_output)
    elif expected_output["parser"] == "SuperAnimalParser":
        check_keypoints_msg(message, expected_output)
    elif expected_output["parser"] == "YuNetParser":
        check_detection_msg(message, expected_output)
    elif expected_output["parser"] == "YOLOExtendedParser":
        check_detection_msg(message, expected_output)
    elif expected_output["parser"] == "DetectionParser":
        check_detection_msg(message, expected_output)



import numpy as np
import pytest

from depthai_nodes import SegmentationMask
from depthai_nodes.message.creators import create_segmentation_message

MASK = np.random.randint(0, 256, (480, 640), dtype=np.int16)


def test_valid_input():
    message = create_segmentation_message(MASK)

    assert isinstance(message, SegmentationMask)
    assert message.mask.shape == (480, 640)
    assert message.mask.dtype == np.int16


def test_invalid_type():
    with pytest.raises(ValueError):
        create_segmentation_message(MASK.tolist())


def test_invalid_shape():
    mask = np.random.randint(0, 256, (480, 640, 3), dtype=np.int16)
    with pytest.raises(ValueError):
        create_segmentation_message(mask)


def test_invalid_dtype():
    with pytest.raises(ValueError):
        create_segmentation_message(MASK.astype(np.uint8))


def test_empty_mask():
    mask = np.empty((0, 0), dtype=np.int16)
    message = create_segmentation_message(mask)

    assert isinstance(message, SegmentationMask)
    assert message.mask.shape == (0, 0)
    assert message.mask.dtype == np.int16



import pytest

from depthai_nodes import Cluster, Clusters
from depthai_nodes.message.creators import (
    create_cluster_message,
)


def test_valid_input():
    clusters = [[[1.0, 2.0], [3.0, 4.0]], [[5.0, 6.0], [7.0, 8.0]]]
    message = create_cluster_message(clusters)

    assert isinstance(message, Clusters)
    assert len(message.clusters) == 2
    assert all(isinstance(cluster, Cluster) for cluster in message.clusters)
    assert message.clusters[0].label == 0
    assert message.clusters[1].label == 1
    assert len(message.clusters[0].points) == 2
    assert len(message.clusters[1].points) == 2


def test_invalid_clusters_type():
    with pytest.raises(TypeError):
        create_cluster_message("not a list")


def test_invalid_cluster_type():
    with pytest.raises(TypeError):
        create_cluster_message(["not a list"])


def test_invalid_point_type():
    with pytest.raises(TypeError):
        create_cluster_message([["not a list"]])


def test_invalid_point_length():
    with pytest.raises(ValueError):
        create_cluster_message([[[1.0, 2.0, 3.0]]])


def test_invalid_value_type():
    with pytest.raises(TypeError):
        create_cluster_message([[[1.0, "not a number"]]])


def test_clusters_input_int():
    clusters = [[[1, 2], [3, 4], [5, 6]], [[7, 8], [9, 10], [11, 12]]]
    create_cluster_message(clusters)


def test_clusters_input_mixed():
    clusters = [[[1, 2], [3, 4], [5, 6]], [[7.0, 8.0], [9.0, 10.0], [11.0, 12.0]]]
    create_cluster_message(clusters)


def test_empty_clusters():
    clusters = []
    message = create_cluster_message(clusters)

    assert isinstance(message, Clusters)
    assert len(message.clusters) == 0


def test_empty_cluster():
    clusters = [[]]
    message = create_cluster_message(clusters)

    assert isinstance(message, Clusters)
    assert len(message.clusters) == 1
    assert len(message.clusters[0].points) == 0



from typing import Dict, Optional

import depthai as dai

from depthai_nodes import ImgDetectionExtended, ImgDetectionsExtended
from depthai_nodes.logging import get_logger
from depthai_nodes.node.base_host_node import BaseHostNode


class ImgDetectionsBridge(BaseHostNode):
    """Transforms the dai.ImgDetections to ImgDetectionsExtended object or vice versa.
    Note that conversion from ImgDetectionsExtended to ImgDetection loses information
    about segmentation, keypoints and rotation.

    Attributes
    ----------
    input : dai.ImgDetections or ImgDetectionsExtended
        The input message for the ImgDetections object.
    output : dai.ImgDetections or ImgDetectionsExtended
        The output message of the transformed ImgDetections object.
    """

    def __init__(self) -> None:
        super().__init__()
        self._logger = get_logger()
        self._log = True
        self._ignore_angle = False
        self._label_encoding = {}
        self._logger.debug("ImgDetectionsBridge initialized")

    def setIgnoreAngle(self, ignore_angle: bool) -> bool:
        """Sets whether to ignore the angle of the detections during transformation.

        @param ignore_angle: Whether to ignore the angle of the detections.
        @type ignore_angle: bool
        """
        if not isinstance(ignore_angle, bool):
            raise ValueError("ignore_angle must be a boolean.")
        self._ignore_angle = ignore_angle
        self._logger.debug(f"Ignore angle set to {self._ignore_angle}")

    def setLabelEncoding(self, label_encoding: Dict[int, str]) -> None:
        """Sets the label encoding.

        @param label_encoding: The label encoding with labels as keys and label names as
            values.
        @type label_encoding: Dict[int, str]
        """
        if not isinstance(label_encoding, Dict):
            raise ValueError("label_encoding must be a dictionary.")
        self._label_encoding = label_encoding
        self._logger.debug(f"Label encoding set to {self._label_encoding}")

    def build(
        self,
        msg: dai.Node.Output,
        ignore_angle: bool = False,
        label_encoding: Optional[Dict[int, str]] = None,
    ) -> "ImgDetectionsBridge":
        """Configures the node connections.

        @param msg: The input message for the ImgDetections object.
        @type msg: dai.Node.Output
        @param ignore_angle: Whether to ignore the angle of the detections.
        @type ignore_angle: bool
        @param label_encoding: The label encoding with labels as keys and label names as
            values.
        @type label_encoding: Dict[int, str]
        @return: The node object with the transformed ImgDetections object.
        @rtype: ImgDetectionsBridge
        """
        self.link_args(msg)
        self.setIgnoreAngle(ignore_angle)
        if label_encoding is not None:
            self.setLabelEncoding(label_encoding)
        self._logger.debug(
            f"ImgDetectionsBridge built with ignore_angle={ignore_angle}, label_encoding={label_encoding}"
        )
        return self

    def process(self, msg: dai.Buffer) -> None:
        """Transforms the incoming ImgDetections object.

        @param msg: The input message for the ImgDetections object.
        @type msg: dai.ImgDetections or ImgDetectionsExtended
        """
        self._logger.debug("Processing new input")
        if isinstance(msg, dai.ImgDetections):
            msg_transformed = self._img_det_to_img_det_ext(msg)
        elif isinstance(msg, ImgDetectionsExtended):
            msg_transformed = self._img_det_ext_to_img_det(msg)
            if self._log:
                self._logger.warning(
                    "You are using ImgDetectionsBridge to transform from ImgDetectionsExtended to ImgDetections. This results in lose of keypoint, segmentation and bbox rotation information if present in the original message."
                )
                self._log = False  # only log once
        else:
            raise TypeError(
                f"Expected dai.ImgDetections or ImgDetectionsExtended, got {type(msg)}"
            )

        msg_transformed.setTimestamp(msg.getTimestamp())
        msg_transformed.setSequenceNum(msg.getSequenceNum())
        msg_transformed.setTransformation(msg.getTransformation())

        self._logger.debug("Detection message created")

        self.out.send(msg_transformed)

        self._logger.debug("Message sent successfully")

    def _img_det_to_img_det_ext(
        self, img_dets: dai.ImgDetections
    ) -> ImgDetectionsExtended:
        """Transforms the incoming dai.ImgDetections object to
        ImgDetectionsExtended."."""
        assert isinstance(img_dets, dai.ImgDetections)

        img_dets_ext = ImgDetectionsExtended()

        detections_transformed = []
        for detection in img_dets.detections:
            detection_transformed = ImgDetectionExtended()
            detection_transformed.label = detection.label
            label_name = self._label_encoding.get(detection.label)
            if label_name is not None:
                detection_transformed.label_name = label_name
            detection_transformed.confidence = detection.confidence
            x_center = (detection.xmin + detection.xmax) / 2
            y_center = (detection.ymin + detection.ymax) / 2
            width = detection.xmax - detection.xmin
            height = detection.ymax - detection.ymin
            detection_transformed.rotated_rect = (
                x_center,
                y_center,
                width,
                height,
                0,  # dai.ImgDetections has no angle info
            )
            detections_transformed.append(detection_transformed)

        img_dets_ext.detections = detections_transformed

        return img_dets_ext

    def _img_det_ext_to_img_det(
        self, img_det_ext: ImgDetectionsExtended
    ) -> dai.ImgDetections:
        """Transforms the incoming ImgDetectionsExtended object to
        dai.ImgDetections."."""
        assert isinstance(img_det_ext, ImgDetectionsExtended)

        img_dets = dai.ImgDetections()

        detections_transformed = []
        for detection in img_det_ext.detections:
            detection_transformed = dai.ImgDetection()
            if detection.label >= 0:
                detection_transformed.label = detection.label
            detection_transformed.confidence = detection.confidence
            if not self._ignore_angle and detection.rotated_rect.angle != 0:
                raise NotImplementedError(
                    "Unable to transform ImgDetectionsExtended with rotation."
                )
            xmin, ymin, xmax, ymax = detection.rotated_rect.getOuterRect()
            detection_transformed.xmin = xmin
            detection_transformed.ymin = ymin
            detection_transformed.xmax = xmax
            detection_transformed.ymax = ymax

            detections_transformed.append(detection_transformed)

        img_dets.detections = detections_transformed

        return img_dets



from typing import Tuple

import cv2
import numpy as np


def to_planar(arr: np.ndarray, shape: Tuple) -> np.ndarray:
    """Converts the input image `arr` (NumPy array) to the planar format expected by
    depthai. The image is resized to the dimensions specified in `shape`.

    @param arr: Input NumPy array (image).
    @type arr: np.ndarray
    @param shape: Target dimensions (width, height).
    @type shape: tuple
    @return: A 1D NumPy array with the planar image data.
    @rtype: np.ndarray
    """
    if np.array_equal(arr.shape[:2], shape):
        resized = arr
    else:
        resized = cv2.resize(arr, shape)

    return resized.transpose(2, 0, 1).flatten()



import logging
import time
from collections import deque

import depthai as dai

from tests.utils.constants import LOG_INTERVAL


class QueueMock:
    """Classic queue to imitate the depthai message queue."""

    def __init__(self):
        self._messages = deque()

    def get(self):
        if len(self._messages) == 0:
            return None
        return self._messages.pop()

    def tryGet(self):
        if len(self._messages) == 0:
            return None
        return self._messages.pop()

    def getAll(self):
        messages = list(self._messages)
        self._messages.clear()
        return messages

    def send(self, item):
        self._messages.appendleft(item)

    def __len__(self):
        return len(self._messages)

    def is_empty(self):
        return len(self._messages) == 0


class InfiniteQueueMock(QueueMock):
    """Queue used in InfiniteInput to simulate a long running test.

    We can set the duration for how long the queue will be active and in the end raise
    an Exception that is caught by the parser.
    """

    def __init__(self):
        super().__init__()
        self.duration = 1  # seconds
        self.start_time = time.time()
        self.log_interval = LOG_INTERVAL  # seconds
        self.time_after_last_log = time.time()
        self.log_counter = 1
        self.logger = logging.getLogger(__name__)

    def send(self, item):
        super().send(item)

    def get(self):
        current_time = time.time()
        if current_time - self.start_time > self.duration:
            raise dai.MessageQueue.QueueException

        # Log progress periodically
        elapsed = current_time - self.time_after_last_log
        if elapsed > self.log_interval:
            elapsed = self.log_counter * self.log_interval
            remaining = self.duration - elapsed
            self.logger.info(
                f"Test running... {elapsed:.1f}s elapsed, {remaining:.1f}s remaining"
            )
            self.time_after_last_log = current_time
            self.log_counter += 1
        element = self._messages.pop()
        self.send(element)
        return element

    def tryGet(self):
        current_time = time.time()
        if current_time - self.start_time > self.duration:
            raise dai.MessageQueue.QueueException

        # Log progress periodically
        elapsed = current_time - self.time_after_last_log
        if elapsed > self.log_interval:
            elapsed = self.log_counter * self.log_interval
            remaining = self.duration - elapsed
            print(f"Test running... {elapsed:.1f}s elapsed, {remaining:.1f}s remaining")
            self.time_after_last_log = current_time
            self.log_counter += 1

        if len(self._messages) == 0:
            return None
        element = self._messages.pop()
        self.send(element)
        return element


class OutputQueueMock(QueueMock):
    """Output queue used for checking the output messages from parsers.

    The messages are verified on the fly so the queue doesnt fill up on long running
    tests.
    """

    def __init__(self, checking_function=None, model_slug=None, parser_name=None):
        super().__init__()
        self._checking_function = checking_function
        self._model_slug = model_slug
        self._parser_name = parser_name

    def send(self, item):
        super().send(item)

        # check msg
        if self._checking_function is not None:
            self._checking_function(item, self._model_slug, self._parser_name)
            self._messages.pop()



import numpy as np
import pytest

from depthai_nodes import ImgDetectionExtended, ImgDetectionsExtended
from depthai_nodes.message.creators import (
    create_detection_message,
)

BBOXES = np.array([[0.2, 0.2, 0.4, 0.4], [0.5, 0.5, 0.1, 0.1], [0.1, 0.1, 0.2, 0.2]])
SCORES = np.array([0.1, 0.2, 0.3])
LABELS = np.array([1, 2, 3])
ANGLES = np.array([45, -45, 0])
KPTS = np.array(
    [[[0.1, 0.1], [0.2, 0.2]], [[0.3, 0.3], [0.4, 0.4]], [[0.5, 0.5], [0.5, 0.5]]]
)
KPTS_SCORES = np.array([[0.9, 0.8], [0.7, 0.6], [0.5, 0.4]])
MASKS = np.array([[0, 1], [1, 0], [0, 0]], dtype=np.int16)

ONE_BBOX = np.array([[0.5, 0.5, 0.2, 0.2]])
ONE_SCORE = np.array([0.9])


def test_valid_input():
    message = create_detection_message(
        bboxes=BBOXES,
        scores=SCORES,
        labels=LABELS,
        angles=ANGLES,
        keypoints=KPTS,
        keypoints_scores=KPTS_SCORES,
        masks=MASKS,
    )

    assert isinstance(message, ImgDetectionsExtended)
    assert len(message.detections) == 3

    for i, detection in enumerate(message.detections):
        assert isinstance(detection, ImgDetectionExtended)
        assert detection.confidence == SCORES[i]
        assert detection.label == LABELS[i]
        assert detection.rotated_rect.angle == ANGLES[i]

        assert np.allclose(detection.rotated_rect.center.x, BBOXES[i][0], atol=1e-3)
        assert np.allclose(detection.rotated_rect.center.y, BBOXES[i][1], atol=1e-3)
        assert np.allclose(detection.rotated_rect.size.width, BBOXES[i][2], atol=1e-3)
        assert np.allclose(detection.rotated_rect.size.height, BBOXES[i][3], atol=1e-3)

        for j, keypoint in enumerate(detection.keypoints):
            assert keypoint.x == KPTS[i][j][0]
            assert keypoint.y == KPTS[i][j][1]
            assert keypoint.confidence == KPTS_SCORES[i][j]

    assert np.allclose(message.masks, MASKS, atol=1e-3)


def test_valid_input_no_optional():
    message = create_detection_message(BBOXES, SCORES)

    assert isinstance(message, ImgDetectionsExtended)
    assert len(message.detections) == 3

    for i, detection in enumerate(message.detections):
        assert isinstance(detection, ImgDetectionExtended)
        assert detection.confidence == SCORES[i]


def test_bboxes_scores_labels():
    message = create_detection_message(bboxes=BBOXES, scores=SCORES, labels=LABELS)

    for i, label in enumerate(LABELS):
        assert message.detections[i].label == label


def test_bboxes_scores_keypoints():
    message = create_detection_message(bboxes=BBOXES, scores=SCORES, keypoints=KPTS)

    assert isinstance(message, ImgDetectionsExtended)

    for i, detection in enumerate(message.detections):
        for ii, keypoint in enumerate(detection.keypoints):
            assert keypoint.x == KPTS[i][ii][0]
            assert keypoint.y == KPTS[i][ii][1]


def test_bboxes_masks():
    message = create_detection_message(bboxes=BBOXES, scores=SCORES, masks=MASKS)

    assert isinstance(message, ImgDetectionsExtended)
    assert np.array_equal(message.masks, MASKS)


def test_no_bboxes_masks():
    bboxes = np.array([])
    scores = np.array([])

    message = create_detection_message(bboxes=bboxes, scores=scores, masks=MASKS)

    assert isinstance(message, ImgDetectionsExtended)
    assert np.array_equal(message.masks, MASKS)


def test_no_masks_no_bboxes():
    bboxes = np.array([])
    scores = np.array([])
    mask = np.array([[]], dtype=np.int16)

    message = create_detection_message(bboxes=bboxes, scores=scores, masks=mask)

    assert isinstance(message, ImgDetectionsExtended)
    assert np.array_equal(message.masks, mask)


def test_empty_bboxes():
    bboxes = np.array([])
    scores = np.array([])
    message = create_detection_message(bboxes, scores)

    assert isinstance(message, ImgDetectionsExtended)
    assert len(message.detections) == 0


def test_invalid_bboxes_type():
    with pytest.raises(ValueError):
        create_detection_message(BBOXES.tolist(), SCORES)


def test_invalid_bboxes_shape():
    with pytest.raises(ValueError):
        create_detection_message(ONE_BBOX[0], ONE_SCORE)


def test_invalid_scores_type():
    with pytest.raises(ValueError):
        create_detection_message(BBOXES, SCORES.tolist())


def test_invalid_scores_shape():
    with pytest.raises(ValueError):
        create_detection_message(ONE_BBOX, np.array([ONE_SCORE]))


def test_mismatched_bboxes_scores_length():
    with pytest.raises(ValueError):
        create_detection_message(ONE_BBOX, np.array([0.9, 0.8]))


def test_invalid_labels_type():
    with pytest.raises(ValueError):
        create_detection_message(ONE_BBOX, ONE_SCORE, labels=[1])


def test_mismatched_bboxes_labels_length():
    with pytest.raises(ValueError):
        create_detection_message(ONE_BBOX, ONE_SCORE, labels=np.array([1, 2]))


def test_invalid_angles_type():
    with pytest.raises(ValueError):
        create_detection_message(ONE_BBOX, ONE_SCORE, angles=[45])


def test_mismatched_bboxes_angles_length():
    with pytest.raises(ValueError):
        create_detection_message(
            ONE_BBOX,
            ONE_SCORE,
            angles=np.array([45, -45]),
        )


def test_invalid_angles_range():
    with pytest.raises(ValueError):
        create_detection_message(ONE_BBOX, ONE_SCORE, angles=np.array([400]))


def test_invalid_keypoints_type():
    with pytest.raises(ValueError):
        create_detection_message(ONE_BBOX, ONE_SCORE, keypoints=[[[0.1, 0.1]]])


def test_invalid_keypoints_shape():
    with pytest.raises(ValueError):
        create_detection_message(
            ONE_BBOX,
            ONE_SCORE,
            keypoints=np.array([0.1, 0.1]),
        )


def test_mismatched_bboxes_keypoints_length():
    with pytest.raises(ValueError):
        create_detection_message(
            ONE_BBOX,
            ONE_SCORE,
            keypoints=np.array([[[0.1, 0.1]], [[0.2, 0.2]]]),
        )


def test_invalid_keypoints_dim():
    with pytest.raises(ValueError):
        create_detection_message(
            ONE_BBOX,
            ONE_SCORE,
            keypoints=np.array([[[0.1, 0.1, 0.1, 0.1]]]),
        )


def test_invalid_keypoints_scores_type():
    with pytest.raises(ValueError):
        create_detection_message(
            ONE_BBOX,
            ONE_SCORE,
            keypoints=np.array([[[0.1, 0.1]]]),
            keypoints_scores=[[[0.9]]],
        )


def test_mismatched_keypoints_keypoints_scores_length():
    with pytest.raises(ValueError):
        create_detection_message(
            ONE_BBOX,
            ONE_SCORE,
            keypoints=np.array([[[0.1, 0.1]]]),
            keypoints_scores=np.array([[[0.9], [0.8]]]),
        )


def test_invalid_keypoints_scores_range():
    with pytest.raises(ValueError):
        create_detection_message(
            ONE_BBOX,
            ONE_SCORE,
            keypoints=np.array([[[0.1, 0.1]]]),
            keypoints_scores=np.array([[[1.1]]]),
        )


def test_invalid_masks_type():
    with pytest.raises(ValueError):
        create_detection_message(ONE_BBOX, ONE_SCORE, masks=[[0, 1]])


def test_invalid_masks_shape():
    with pytest.raises(ValueError):
        create_detection_message(ONE_BBOX, ONE_SCORE, masks=np.array([0, 1]))



import time

import depthai as dai
import numpy as np
import pytest

from depthai_nodes.node import TilesPatcher, Tiling
from tests.utils import LOG_INTERVAL, OutputMock, create_img_detection


@pytest.fixture
def duration(request):
    d = request.config.getoption("--duration")
    if d is None:
        return 1e-6
    return d


@pytest.fixture
def patcher():
    return TilesPatcher()


@pytest.fixture
def tile_manager():
    tm = Tiling()
    tm.overlap = 0.1
    tm.grid_size = (2, 2)
    tm.x = np.array([640, 360])
    tm.nn_shape = (300, 300)
    tm.img_shape = (1280, 720)
    tm.tile_positions = [
        {"coords": (0, 0, 640, 360), "scaled_size": (300, 300)},
        {"coords": (640, 360, 1280, 720), "scaled_size": (300, 300)},
    ]
    return tm


@pytest.fixture
def img_detection_1():
    return create_img_detection(
        [0.1, 0.1, 0.2, 0.2],
        1,
        0.9,
    )


@pytest.fixture
def img_detection_2():
    return create_img_detection(
        [0.7, 0.7, 0.8, 0.8],
        2,
        0.8,
    )


def create_dummy_nn_output(detection):
    nn_output = dai.ImgDetections()
    nn_output.detections = [detection]
    return nn_output


def test_build_valid(patcher: TilesPatcher, tile_manager: Tiling):
    patcher.build(
        tile_manager=tile_manager,
        nn=OutputMock(),
        conf_thresh=0.5,
        iou_thresh=0.6,
    )

    assert patcher.conf_thresh == 0.5
    assert patcher.iou_thresh == 0.6
    assert patcher.tile_manager is tile_manager
    assert patcher.expected_tiles_count == len(tile_manager.tile_positions)


def test_process_accumulation(
    patcher: TilesPatcher,
    tile_manager: Tiling,
    img_detection_1: dai.ImgDetection,
    img_detection_2: dai.ImgDetection,
    duration: float,
):
    patcher.build(tile_manager=tile_manager, nn=OutputMock())

    nn_output1 = create_dummy_nn_output(img_detection_1)
    nn_output2 = create_dummy_nn_output(img_detection_2)

    out_q = patcher.out.createOutputQueue()

    start_time = time.time()

    last_log_time = time.time()
    while time.time() - start_time < duration:
        if time.time() - last_log_time > LOG_INTERVAL:
            print(
                f"Test running... {time.time()-start_time:.1f}s elapsed, {duration-time.time()+start_time:.1f}s remaining"
            )
            last_log_time = time.time()
        patcher.process(nn_output1)
        assert out_q.is_empty()
        patcher.process(nn_output2)
        assert len(out_q._messages) == 1
        detections_msg = out_q.get()
        assert out_q.is_empty()
        assert isinstance(detections_msg, dai.ImgDetections)
        assert len(detections_msg.detections) == 2



from typing import List, Optional, Tuple

import depthai as dai

from .input import InputMock
from .node import NodeMock
from .output import OutputMock
from .pipeline import PipelineMock


class HostNodeMock(NodeMock):
    """Mock class for the depthai HostNode.

    The class is used to create a mock pipeline for testing purposes.
    """

    def __init__(self):
        super().__init__()
        self._output = OutputMock()
        self._parent_pipeline = None
        self._input = InputMock()
        self._linked_args: Optional[Tuple[OutputMock, ...]] = None
        self._sendProcessingToPipeline = False
        self._pipeline = PipelineMock()

    def link_args(self, *args):
        for arg in args:
            assert isinstance(arg, OutputMock)
        self._linked_args = args

    @property
    def out(self):
        return self._output

    @property
    def output(self):
        return self._output

    @output.setter
    def output(self, output):
        self._output = output

    def createOutput(self, possibleDatatypes: List[Tuple[dai.DatatypeEnum, bool]]):
        return self._output

    def sendProcessingToPipeline(self, send: bool):
        self._sendProcessingToPipeline = send

    def getParentPipeline(self):
        return self._pipeline



from datetime import timedelta
from typing import List

import depthai as dai
import numpy as np

import depthai_nodes.message.creators as creators
from depthai_nodes import ImgDetectionExtended

from .constants import ARRAYS, DETECTIONS


def create_img_detection(
    bbox: List[float] = DETECTIONS["bboxes"][0],
    label: int = DETECTIONS["labels"][0],
    score: float = DETECTIONS["scores"][0],
):
    """Creates a dai.ImgDetection object.

    @param det: Detection dict with keys "bbox" ([xmin, ymin, xmax, ymax]), "label"
        (int), and "confidence" (float).
    @type det: dict
    @return: The created dai.ImgDetection object.
    @rtype: dai.ImgDetection
    """

    img_det = dai.ImgDetection()
    img_det.xmin, img_det.ymin, img_det.xmax, img_det.ymax = bbox
    img_det.label = label
    img_det.confidence = score
    return img_det


def create_img_detections(
    bboxs: np.ndarray = DETECTIONS["bboxes"],
    labels: np.ndarray = DETECTIONS["labels"],
    scores: np.ndarray = DETECTIONS["scores"],
    timestamp: int = timedelta(days=1, hours=1, minutes=1, seconds=1, milliseconds=0),
):
    """Creates a dai.ImgDetections object.

    @param dets: List of detection dicts, each containing "bbox" ([xmin, ymin, xmax,
        ymax]), "label" (int), and "confidence" (float).
    @type dets: List[Dict]
    @return: The created dai.ImgDetections object.
    @rtype: dai.ImgDetections
    """
    img_dets = dai.ImgDetections()
    img_dets.detections = [
        create_img_detection(bbox.tolist(), label.item(), score.item())
        for bbox, label, score in zip(bboxs, labels, scores)
    ]
    img_dets.setTimestamp(timestamp)
    return img_dets


def create_img_detection_extended(
    bbox: np.ndarray = DETECTIONS["bboxes"][0],
    label: np.ndarray = DETECTIONS["labels"][0],
    score: np.ndarray = DETECTIONS["scores"][0],
):
    """Creates a ImgDetectionExtended object.

    @param det: Detection dict with keys "bbox" ([xmin, ymin, xmax, ymax]), "label"
        (int), and "confidence" (float).
    @type det: dict
    @return: The created ImgDetectionExtended object.
    @rtype: ImgDetectionExtended
    """
    img_det_ext = ImgDetectionExtended()
    xmin, ymin, xmax, ymax = bbox.tolist()
    x_center = (xmin + xmax) / 2
    y_center = (ymin + ymax) / 2
    width = xmax - xmin
    height = ymax - ymin
    img_det_ext.label = label.item()
    img_det_ext.confidence = score.item()
    img_det_ext.rotated_rect = (x_center, y_center, width, height, 0)
    return img_det_ext


def create_img_detections_extended(
    bboxes: np.ndarray = DETECTIONS["bboxes"],
    scores: np.ndarray = DETECTIONS["scores"],
    angles: np.ndarray = DETECTIONS["angles"],
    labels: np.ndarray = DETECTIONS["labels"],
    keypoints: np.ndarray = DETECTIONS["keypoints"],
    keypoints_scores: np.ndarray = DETECTIONS["keypoints_scores"],
    masks: np.ndarray = ARRAYS["2d"],
):
    return creators.create_detection_message(
        bboxes=bboxes,
        scores=scores,
        angles=angles,
        labels=labels,
        keypoints=keypoints,
        keypoints_scores=keypoints_scores,
        masks=masks,
    )



import depthai as dai
import numpy as np
import pytest

from depthai_nodes import Map2D


@pytest.fixture
def map2d():
    return Map2D()


def test_map2d_initialization(map2d: Map2D):
    assert np.array_equal(map2d.map, np.array([]))
    assert map2d.width is None
    assert map2d.height is None
    assert map2d.transformation is None


def test_map2d_set_map(map2d: Map2D):
    map_array = np.random.rand(480, 640).astype(np.float32)
    map2d.map = map_array
    assert np.array_equal(map2d.map, map_array)
    assert map2d.width == 640
    assert map2d.height == 480

    with pytest.raises(TypeError):
        map2d.map = "not a numpy array"

    with pytest.raises(ValueError):
        map2d.map = np.random.rand(480, 640, 3).astype(np.float32)

    with pytest.raises(ValueError):
        map2d.map = np.random.rand(480, 640).astype(np.float64)


def test_map2d_set_transformation(map2d: Map2D):
    transformation = dai.ImgTransformation()
    map2d.transformation = transformation
    assert map2d.transformation == transformation

    with pytest.raises(TypeError):
        map2d.transformation = "not a dai.ImgTransformation"


def test_map2d_set_transformation_none(map2d: Map2D):
    map2d.transformation = None
    assert map2d.transformation is None



import numpy as np

import depthai_nodes.message.creators as creators

from .constants import DETECTIONS


def create_keypoints(
    keypoints: np.ndarray = DETECTIONS["keypoints"][0],
    scores: np.ndarray = DETECTIONS["keypoints_scores"][0],
):
    return creators.create_keypoints_message(keypoints=keypoints, scores=scores)



import os

import depthai as dai


def get_calibration_handler():
    calibration_handler_json_path = os.path.join(
        os.path.dirname(__file__), "calib.json"
    )
    calibration_handler = dai.CalibrationHandler(calibration_handler_json_path)
    return calibration_handler



class RequireSuperInit(type):
    """Checking whether every class in the hierarchy calls super().__init__() inside
    their __init__ method."""

    def __call__(cls, *args, **kwargs):
        inst = cls.__new__(cls, *args, **kwargs)
        inst._b_init_called = False
        cls.__init__(inst, *args, **kwargs)

        if not getattr(inst, "_b_init_called", False):
            raise RuntimeError(
                f"{cls.__name__}.__init__ didn't call super().__init__()!"
            )
        return inst


class NodeMock(metaclass=RequireSuperInit):
    def __init__(self):
        self._b_init_called = True  # the metaclass catches if it wasn’t set during construction (i.e. if one of the child classes didn’t call super().__init__())



from typing import List, Union

import depthai_nodes.message.creators as creators

from .constants import COLLECTIONS


def create_clusters(
    clusters: List[List[List[Union[float, int]]]] = COLLECTIONS["clusters"],
):
    return creators.create_cluster_message(clusters=clusters)



import depthai as dai
import pytest

from depthai_nodes.node import ParsingNeuralNetwork
from tests.utils import InputMock, NeuralNetworkMock, PipelineMock


@pytest.fixture
def pipeline():
    dai.node.NeuralNetwork = NeuralNetworkMock
    return PipelineMock()


def validate(pnn: ParsingNeuralNetwork, nn_archive: dai.NNArchive):
    num_heads = len(nn_archive.getConfig().model.heads)

    assert (
        len(pnn._parsers) == num_heads
    ), f"Expected {num_heads} parsers, got {len(pnn._parsers)}"

    for parser_ix in range(num_heads):
        p = pnn.getParser(parser_ix)
        assert p is not None, f"Parser {parser_ix} is None"

        output = pnn.getOutput(parser_ix)
        assert output is not None, f"Output {parser_ix} is None"


@pytest.mark.parametrize(
    "model",
    [
        "luxonis/yunet:320x240",
        "luxonis/vehicle-attributes-classification:72x72",
        "luxonis/mediapipe-hand-landmarker:224x224",
        "luxonis/yolov6-nano:r2-coco-512x288",
        "luxonis/mobilenet-ssd:300x300",
    ],
)
def test_parsing_neural_network(pipeline: PipelineMock, model: str):
    nn_archive = dai.NNArchive(
        dai.getModelFromZoo(dai.NNModelDescription(model, "RVC2"))
    )
    pnn: ParsingNeuralNetwork = pipeline.create(ParsingNeuralNetwork).build(
        input=InputMock(), nn_source=nn_archive, fps=30
    )

    validate(pnn, nn_archive)


def test_parsing_neural_network_nn_archive(pipeline: PipelineMock):
    model = "luxonis/yunet:320x240"
    nn_archive = dai.NNArchive(
        dai.getModelFromZoo(dai.NNModelDescription(model, "RVC2"))
    )
    pnn: ParsingNeuralNetwork = pipeline.create(ParsingNeuralNetwork).build(
        input=InputMock(), nn_source=nn_archive, fps=30
    )

    validate(pnn, nn_archive)

    model = "luxonis/vehicle-attributes-classification:72x72"
    nn_archive = dai.NNArchive(
        dai.getModelFromZoo(dai.NNModelDescription(model, "RVC2"))
    )
    pnn.setNNArchive(nn_archive)

    validate(pnn, nn_archive)



import cv2
import depthai as dai

from depthai_nodes.node.base_host_node import BaseHostNode


class ImgFrameOverlay(BaseHostNode):
    """A host node that receives two dai.ImgFrame objects and overlays them into a
    single one.

    Attributes
    ----------
    frame1 : dai.ImgFrame
        The input message for the background frame.
    frame2 : dai.ImgFrame
        The input message for the foreground frame.
    alpha: float
        The weight of the background frame in the overlay. By default, the weight is 0.5
            which means that both frames are represented equally in the overlay.
    out : dai.ImgFrame
        The output message for the overlay frame.
    """

    def __init__(self, alpha: float = 0.5) -> None:
        super().__init__()
        self.setAlpha(alpha)
        self._logger.debug(f"ImgFrameOverlay initialized with alpha={alpha}")

    def setAlpha(self, alpha: float) -> None:
        """Sets the alpha.

        @param alpha: The weight of the background frame in the overlay.
        @type alpha: float
        """
        if not isinstance(alpha, float):
            raise ValueError("Alpha must be a float")
        if not 0.0 <= alpha <= 1.0:
            raise ValueError("Alpha must be between 0.0 and 1.0")
        self._alpha = alpha
        self._logger.debug(f"Alpha set to {self._alpha}")

    def build(
        self, frame1: dai.Node.Output, frame2: dai.Node.Output, alpha: float = None
    ) -> "ImgFrameOverlay":
        """Configures the node connections.

        @param frame1: The input message for the background frame.
        @type frame1: dai.Node.Output
        @param frame2: The input message for the foreground frame.
        @type frame2: dai.Node.Output
        @param alpha: The weight of the background frame in the overlay.
        @type alpha: float
        @return: The node object with the background and foreground streams overlaid.
        @rtype: ImgFrameOverlay
        """
        self.link_args(frame1, frame2)

        if alpha is not None:
            self.setAlpha(alpha)

        self._logger.debug(f"ImgFrameOverlay built with alpha={alpha}")

        return self

    def process(self, frame1: dai.Buffer, frame2: dai.Buffer) -> None:
        """Processes incoming frames and overlays them.

        @param frame1: The input message for the background frame.
        @type frame1: dai.ImgFrame
        @param frame2: The input message for the foreground frame.
        @type frame2: dai.ImgFrame
        """
        self._logger.debug("Processing new input")
        assert isinstance(frame1, dai.ImgFrame)
        assert isinstance(frame2, dai.ImgFrame)

        background_frame = frame1.getCvFrame()
        foreground_frame = frame2.getCvFrame()

        # reshape foreground to match the background shape
        foreground_frame = cv2.resize(
            foreground_frame,
            dsize=(background_frame.shape[1], background_frame.shape[0]),
            interpolation=cv2.INTER_LINEAR,
        )

        overlay_frame = cv2.addWeighted(
            background_frame,
            self._alpha,
            foreground_frame,
            1 - self._alpha,
            0,
        )

        overlay = dai.ImgFrame()
        overlay.setCvFrame(
            overlay_frame,
            self._img_frame_type,
        )
        overlay.setTimestamp(frame1.getTimestamp())
        overlay.setSequenceNum(frame1.getSequenceNum())

        self._logger.debug("ImgFrame message created")

        self.out.send(overlay)

        self._logger.debug("Message sent successfully")



import cv2
import depthai as dai
import numpy as np

from depthai_nodes.message import ImgDetectionsExtended, Map2D, SegmentationMask
from depthai_nodes.message.utils import copy_message
from depthai_nodes.node.base_host_node import BaseHostNode


class ApplyColormap(BaseHostNode):
    """A host node that applies a colormap to a given 2D array (e.g. depth maps,
    segmentation masks, heatmaps, etc.).

    Attributes
    ----------
    colormap_value : Optional[int]
        OpenCV colormap enum value. Determines the applied color mapping.
    max_value : Optional[int]
        Maximum value to consider for normalization. If set lower than the map's actual maximum, the map's maximum will be used instead.
    instance_to_semantic_mask : Optional[bool]
        If True, converts instance segmentation masks to semantic segmentation masks. Note that this is only relevant for ImgDetectionsExtended messages.
    arr : dai.ImgFrame or Map2D or ImgDetectionsExtended
        The input message with a 2D array.
    output : dai.ImgFrame
        The output message for a colorized frame.
    """

    def __init__(
        self,
        colormap_value: int = cv2.COLORMAP_JET,
        max_value: int = 0,
        instance_to_semantic_mask: bool = False,
    ) -> None:
        super().__init__()

        self.out.setPossibleDatatypes([(dai.DatatypeEnum.ImgFrame, True)])

        self.setColormap(colormap_value)
        self.setMaxValue(max_value)
        self.setInstanceToSemanticMask(instance_to_semantic_mask)

        self._logger.debug(
            f"ApplyColormap initialized with colormap_value={colormap_value}, max_value={max_value}, instance_to_semantic_mask={instance_to_semantic_mask}",
        )

    def setColormap(self, colormap_value: int) -> None:
        """Sets the applied color mapping.

        @param colormap_value: OpenCV colormap enum value (e.g. cv2.COLORMAP_HOT)
        @type colormap_value: int
        """
        if not isinstance(colormap_value, int):
            raise ValueError("colormap_value must be an integer.")
        colormap = cv2.applyColorMap(np.arange(256, dtype=np.uint8), colormap_value)
        colormap[0] = [0, 0, 0]  # Set zero values to black
        self._colormap = colormap
        self._logger.debug(f"Colormap set to {self._colormap}")

    def setMaxValue(self, max_value: int) -> None:
        """Sets the maximum frame value for normalization.

        @param max_value: Maximum frame value.
        @type max_value: int
        """
        if not isinstance(max_value, int):
            raise ValueError("max_value must be an integer.")
        self._max_value = max_value
        self._logger.debug(f"Max value set to {self._max_value}")

    def setInstanceToSemanticMask(self, instance_to_semantic_mask: bool) -> None:
        """Sets the instance to semantic mask flag.

        @param instance_to_semantic_mask: If True, converts instance segmentation masks
            to semantic segmentation masks.
        @type instance_to_semantic_mask: bool
        """
        if not isinstance(instance_to_semantic_mask, bool):
            raise ValueError("instance_to_semantic_mask must be a boolean.")
        self._instance_to_semantic_mask = instance_to_semantic_mask
        self._logger.debug(
            f"Instance to semantic mask set to {self._instance_to_semantic_mask}"
        )

    def build(self, arr: dai.Node.Output) -> "ApplyColormap":
        """Configures the node connections.

        @param frame: Output with 2D array.
        @type frame: depthai.Node.Output
        @return: The node object with input stream connected
        @rtype: ApplyColormap
        """
        self.link_args(arr)
        self._logger.debug("ApplyColormap built")
        return self

    def process(self, msg: dai.Buffer) -> None:
        """Processes incoming 2D arrays and converts them to colored frames.

        @param msg: The input message with a 2D array.
        @type msg: dai.ImgFrame or Map2D or ImgDetectionsExtended
        @param instance_to_semantic_segmentation: If True, converts instance
            segmentation masks to semantic segmentation masks.
        @type instance_to_semantic_segmentation: bool
        """
        self._logger.debug("Processing new input")
        msg_copy = copy_message(msg)

        if isinstance(msg, SegmentationMask):
            arr = msg_copy.mask
        elif isinstance(msg, dai.ImgFrame):
            if not msg.getType().name.startswith("RAW"):
                raise TypeError(f"Expected image type RAW, got {msg.getType().name}")
            arr = msg.getCvFrame()
        elif isinstance(msg, Map2D):
            arr = msg_copy.map
        elif isinstance(msg, ImgDetectionsExtended):
            if self._instance_to_semantic_mask:
                labels = {
                    idx: detection.label for idx, detection in enumerate(msg.detections)
                }
                labels[-1] = -1  # background class
                arr = np.vectorize(lambda x: labels.get(x, -1))(
                    msg_copy.masks  # instance segmentation mask
                )  # semantic segmentation mask
            else:
                arr = msg_copy.masks  # semantic segmentation mask
        else:
            raise ValueError(f"Unsupported input type {type(msg)}.")

        # make sure that min value == 0 to ensure proper normalization
        arr += np.abs(arr.min()) if arr.min() < 0 else 0

        max_value = max(self._max_value, arr.max())
        if max_value == 0:
            color_arr = np.zeros(
                (arr.shape[0], arr.shape[1], 3),
                dtype=np.uint8,
            )
        else:
            color_arr = cv2.applyColorMap(
                ((arr / max_value) * 255).astype(np.uint8),
                self._colormap,
            )

        frame = dai.ImgFrame()
        frame.setCvFrame(
            color_arr,
            self._img_frame_type,
        )
        frame.setTimestamp(msg.getTimestamp())
        frame.setSequenceNum(msg.getSequenceNum())

        self._logger.debug("ImgFrame message created")

        self.out.send(frame)

        self._logger.debug("Message sent successfully")



import math
import time
from typing import Union

import depthai as dai
import pytest
from pytest import FixtureRequest

from depthai_nodes import ImgDetectionExtended, ImgDetectionsExtended
from depthai_nodes.node import ImgDetectionsBridge
from tests.utils import (
    LOG_INTERVAL,
    OutputMock,
    create_img_detections,
    create_img_detections_extended,
)

img_det_types = ["img_detections", "img_detections_extended"]


@pytest.fixture(scope="session")
def duration(request):
    d = request.config.getoption("--duration")
    if d is None:
        return 1e-6
    return d


@pytest.fixture
def bridge():
    return ImgDetectionsBridge()


@pytest.fixture
def img_detections():
    return create_img_detections()


@pytest.fixture
def img_detections_extended():
    return create_img_detections_extended()


def test_building(bridge: ImgDetectionsBridge):
    bridge.build(OutputMock())


@pytest.mark.parametrize(
    "img_detections_type",
    img_det_types,
)
def test_processing(
    bridge: ImgDetectionsBridge,
    request: FixtureRequest,
    img_detections_type: str,
    duration: float,
):
    dets: Union[ImgDetectionsExtended, dai.ImgDetections] = request.getfixturevalue(
        img_detections_type
    )

    o_dets = OutputMock()
    bridge.build(o_dets, ignore_angle=True)
    q_dets = o_dets.createOutputQueue()
    q_dets_transformed = bridge.out.createOutputQueue()

    modified_duration = duration / len(img_det_types)

    def _identical_detections(
        img_dets: dai.ImgDetections, img_dets_ext: ImgDetectionExtended
    ):
        assert isinstance(img_dets, dai.ImgDetections)
        assert isinstance(img_dets_ext, ImgDetectionsExtended)
        assert len(img_dets.detections) == len(img_dets_ext.detections)
        for img_det, img_det_ext in zip(img_dets.detections, img_dets_ext.detections):
            xmin, ymin, xmax, ymax = img_det_ext.rotated_rect.getOuterRect()
            assert math.isclose(img_det.xmin, xmin, rel_tol=1e-6)
            assert math.isclose(img_det.ymin, ymin, rel_tol=1e-6)
            assert math.isclose(img_det.xmax, xmax, rel_tol=1e-6)
            assert math.isclose(img_det.ymax, ymax, rel_tol=1e-6)
            assert math.isclose(img_det.label, img_det_ext.label, rel_tol=1e-6)
            assert math.isclose(
                img_det.confidence, img_det_ext.confidence, rel_tol=1e-6
            )

    start_time = time.time()
    last_log_time = time.time()
    while time.time() - start_time < modified_duration:
        if time.time() - last_log_time > LOG_INTERVAL:
            print(
                f"Test running... {time.time()-start_time:.1f}s elapsed, {modified_duration-time.time()+start_time:.1f}s remaining"
            )
            last_log_time = time.time()
        q_dets.send(dets)
        bridge.process(q_dets.get())
        dets_transformed = q_dets_transformed.get()

        if isinstance(dets, dai.ImgDetections):
            _identical_detections(dets, dets_transformed)
        elif isinstance(dets, ImgDetectionsExtended):
            _identical_detections(dets_transformed, dets)
        else:
            raise TypeError(f"Unexpected output message type: {type(dets)}")



from typing import List

from depthai_nodes import Prediction, Predictions


def create_regression_message(predictions: List[float]) -> Predictions:
    """Create a DepthAI message for prediction models.

    @param predictions: Predicted value(s).
    @type predictions: List[float]
    @return: Predictions message containing the predicted value(s).
    @rtype: Predictions
    @raise ValueError: If predictions is not a list.
    @raise ValueError: If each prediction is not a float.
    """

    if not isinstance(predictions, list):
        raise ValueError(f"Predictions should be list, got {type(predictions)}.")

    for prediction in predictions:
        if not isinstance(prediction, float):
            raise ValueError(
                f"Each prediction should be a float, got {type(prediction)} instead."
            )

    prediction_objects_list = []
    for prediction in predictions:
        prediction_object = Prediction()
        prediction_object.prediction = prediction
        prediction_objects_list.append((prediction_object))

    predictions_message = Predictions()
    predictions_message.predictions = prediction_objects_list

    return predictions_message



from typing import List, Union

import depthai as dai

from depthai_nodes import Cluster, Clusters


def create_cluster_message(clusters: List[List[List[Union[float, int]]]]) -> Clusters:
    """Create a DepthAI message for clusters.

    @param clusters: List of clusters. Each cluster is a list of points with x and y
        coordinates.
    @type clusters: List[List[List[Union[float, int]]]]
    @return: Clusters message containing the detected clusters.
    @rtype: Clusters
    @raise TypeError: If the clusters are not a list.
    @raise TypeError: If each cluster is not a list.
    @raise TypeError: If each point is not a list.
    @raise TypeError: If each value in the point is not an int or float.
    """

    if not isinstance(clusters, list):
        raise TypeError(f"clusters must be a list, got {type(clusters)}")
    for cluster in clusters:
        if not isinstance(cluster, list):
            raise TypeError(f"All clusters must be of type List, got {type(cluster)}")
        for point in cluster:
            if not isinstance(point, tuple) and not isinstance(point, list):
                raise TypeError(
                    f"All points in clusters must be of type tuple or list, got {type(point)}"
                )
            if len(point) != 2:
                raise ValueError(f"Each point must have 2 values, got {len(point)}")
            for value in point:
                if not isinstance(value, (float, int)):
                    raise TypeError(
                        f"All items in points must be of type int or float, got {type(value)}"
                    )

    message = Clusters()
    temp = []
    for i, cluster in enumerate(clusters):
        temp_cluster = Cluster()
        temp_cluster.label = i
        temp_cluster.points = [
            dai.Point2f(float(point[0]), float(point[1])) for point in cluster
        ]

        temp.append(temp_cluster)

    message.clusters = temp

    return message



from typing import Any, Dict, Optional, Tuple

import depthai as dai
import numpy as np

from depthai_nodes.message.creators import create_tracked_features_message
from depthai_nodes.node.parsers.base_parser import BaseParser
from depthai_nodes.node.parsers.utils.xfeat import detect_and_compute, match


class XFeatBaseParser(BaseParser):
    """Base parser class for parsing the output of the XFeat model. It is the parent
    class of the XFeatMonoParser and XFeatStereoParser classes.

    Attributes
    ----------
    reference_input : Node.Input
        Reference input for stereo mode. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    target_input : Node.Input
        Target input for stereo mode. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    output_layer_feats : str
        Name of the output layer containing features.
    output_layer_keypoints : str
        Name of the output layer containing keypoints.
    output_layer_heatmaps : str
        Name of the output layer containing heatmaps.
    original_size : Tuple[float, float]
        Original image size.
    input_size : Tuple[float, float]
        Input image size.
    max_keypoints : int
        Maximum number of keypoints to keep.

    Error Handling
    --------------
    **ValueError**: If the number of output layers is not E{3}.
    **ValueError**: If the original image size is not specified.
    **ValueError**: If the input image size is not specified.
    **ValueError**: If the maximum number of keypoints is not specified.
    **ValueError**: If the output layer containing features is not specified.
    **ValueError**: If the output layer containing keypoints is not specified.
    **ValueError**: If the output layer containing heatmaps is not specified.
    """

    def __init__(
        self,
        output_layer_feats: str = "",
        output_layer_keypoints: str = "",
        output_layer_heatmaps: str = "",
        original_size: Tuple[float, float] = None,
        input_size: Tuple[float, float] = (640, 352),
        max_keypoints: int = 4096,
    ) -> None:
        """Initializes the parser node."""
        super().__init__()
        self._target_input = self.createInput()  # used in stereo mode

        self.output_layer_feats = output_layer_feats
        self.output_layer_keypoints = output_layer_keypoints
        self.output_layer_heatmaps = output_layer_heatmaps
        self.original_size = original_size
        self.input_size = input_size
        self.max_keypoints = max_keypoints

        self._logger.debug(
            f"XFeatBaseParser initialized with output_layer_feats='{output_layer_feats}', output_layer_keypoints='{output_layer_keypoints}', output_layer_heatmaps='{output_layer_heatmaps}', original_size={original_size}, input_size={input_size}, max_keypoints={max_keypoints}"
        )

    @property
    def reference_input(self) -> Optional[dai.Node.Input]:
        """Returns the reference input."""
        return self.input

    @property
    def target_input(self) -> Optional[dai.Node.Input]:
        """Returns the target input."""
        return self._target_input

    @reference_input.setter
    def reference_input(self, reference_input: Optional[dai.Node.Input]):
        """Sets the reference input."""
        self.input = reference_input

    @target_input.setter
    def target_input(self, target_input: Optional[dai.Node.Input]):
        """Sets the target input."""
        self._target_input = target_input

    def setOutputLayerFeats(self, output_layer_feats: str) -> None:
        """Sets the output layer containing features.

        @param output_layer_feats: Name of the output layer containing features.
        @type output_layer_feats: str
        """
        if not isinstance(output_layer_feats, str):
            raise ValueError("Output layer containing features must be a string!")
        self.output_layer_feats = output_layer_feats
        self._logger.debug(
            f"Output layer containing features set to '{self.output_layer_feats}'"
        )

    def setOutputLayerKeypoints(self, output_layer_keypoints: str) -> None:
        """Sets the output layer containing keypoints.

        @param output_layer_keypoints: Name of the output layer containing keypoints.
        @type output_layer_keypoints: str
        """
        if not isinstance(output_layer_keypoints, str):
            raise ValueError("Output layer containing keypoints must be a string!")
        self.output_layer_keypoints = output_layer_keypoints
        self._logger.debug(
            f"Output layer containing keypoints set to '{self.output_layer_keypoints}'"
        )

    def setOutputLayerHeatmaps(self, output_layer_heatmaps: str) -> None:
        """Sets the output layer containing heatmaps.

        @param output_layer_heatmaps: Name of the output layer containing heatmaps.
        @type output_layer_heatmaps: str
        """
        if not isinstance(output_layer_heatmaps, str):
            raise ValueError("Output layer containing heatmaps must be a string!")
        self.output_layer_heatmaps = output_layer_heatmaps
        self._logger.debug(
            f"Output layer containing heatmaps set to '{self.output_layer_heatmaps}'"
        )

    def setOriginalSize(self, original_size: Tuple[int, int]) -> None:
        """Sets the original image size.

        @param original_size: Original image size.
        @type original_size: Tuple[int, int]
        """
        if not isinstance(original_size, tuple) or len(original_size) != 2:
            raise ValueError("Original image size must be a tuple of two ints!")
        for size in original_size:
            if not isinstance(size, int):
                raise ValueError("Original image size must be a tuple of two ints!")
        self.original_size = original_size
        self._logger.debug(f"Original image size set to {self.original_size}")

    def setInputSize(self, input_size: Tuple[int, int]) -> None:
        """Sets the input image size.

        @param input_size: Input image size.
        @type input_size: Tuple[int, int]
        """
        if not isinstance(input_size, tuple) or len(input_size) != 2:
            raise ValueError("Input image size must be a tuple of two ints!")
        for size in input_size:
            if not isinstance(size, int):
                raise ValueError("Input image size must be a tuple of two ints!")
        self.input_size = input_size
        self._logger.debug(f"Input image size set to {self.input_size}")

    def setMaxKeypoints(self, max_keypoints: int) -> None:
        """Sets the maximum number of keypoints to keep.

        @param max_keypoints: Maximum number of keypoints.
        @type max_keypoints: int
        """
        if not isinstance(max_keypoints, int):
            raise ValueError("Maximum number of keypoints must be an int!")
        self.max_keypoints = max_keypoints
        self._logger.debug(f"Maximum number of keypoints set to {self.max_keypoints}")

    def build(
        self,
        head_config: Dict[str, Any],
    ) -> "XFeatBaseParser":
        """Configures the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: XFeatBaseParser
        """

        output_layers = head_config.get("outputs", [])
        if len(output_layers) != 3:
            raise ValueError(
                f"Only three output layers supported for XFeat, got {len(output_layers)} layers."
            )

        for layer in output_layers:
            if "feats" in layer:
                self.output_layer_feats = layer
            elif "keypoints" in layer:
                self.output_layer_keypoints = layer
            elif "heatmaps" in layer:
                self.output_layer_heatmaps = layer

        self.original_size = head_config.get("original_size", self.original_size)
        self.input_size = head_config.get("input_size", self.input_size)
        self.max_keypoints = head_config.get("max_keypoints", self.max_keypoints)

        self._logger.debug(
            f"XFeatBaseParser built with output_layer_feats='{self.output_layer_feats}', output_layer_keypoints='{self.output_layer_keypoints}', output_layer_heatmaps='{self.output_layer_heatmaps}', original_size={self.original_size}, input_size={self.input_size}, max_keypoints={self.max_keypoints}"
        )

        return self

    def validateParams(self) -> None:
        """Validates the parameters."""
        if self.original_size is None:
            raise ValueError("Original image size must be specified!")
        if self.input_size is None:
            raise ValueError("Input image size must be specified!")
        if self.max_keypoints is None:
            raise ValueError("Maximum number of keypoints must be specified!")
        if self.output_layer_feats == "":
            raise ValueError("Output layer containing features must be specified!")
        if self.output_layer_keypoints == "":
            raise ValueError("Output layer containing keypoints must be specified!")
        if self.output_layer_heatmaps == "":
            raise ValueError("Output layer containing heatmaps must be specified!")

    def extractTensors(
        self, output: dai.NNData
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Extracts the tensors from the output. It returns the features, keypoints, and
        heatmaps. It also handles the reshaping of the tensors by requesting the NCHW
        storage order.

        @param output: Output from the Neural Network node.
        @type output: dai.NNData
        @return: Tuple of features, keypoints, and heatmaps.
        @rtype: Tuple[np.ndarray, np.ndarray, np.ndarray]
        """
        feats = output.getTensor(
            self.output_layer_feats,
            dequantize=True,
            storageOrder=dai.TensorInfo.StorageOrder.NCHW,
        ).astype(np.float32)
        keypoints = output.getTensor(
            self.output_layer_keypoints,
            dequantize=True,
            storageOrder=dai.TensorInfo.StorageOrder.NCHW,
        ).astype(np.float32)
        heatmaps = output.getTensor(
            self.output_layer_heatmaps,
            dequantize=True,
            storageOrder=dai.TensorInfo.StorageOrder.NCHW,
        ).astype(np.float32)

        return feats, keypoints, heatmaps


class XFeatMonoParser(XFeatBaseParser):
    """Parser class for parsing the output of the XFeat model. It can be used for
    parsing the output from one source (e.g. one camera). The reference frame can be set
    with trigger method.

    Attributes
    ----------
    output_layer_feats : str
        Name of the output layer containing features.
    output_layer_keypoints : str
        Name of the output layer containing keypoints.
    output_layer_heatmaps : str
        Name of the output layer containing heatmaps.
    original_size : Tuple[float, float]
        Original image size.
    input_size : Tuple[float, float]
        Input image size.
    max_keypoints : int
        Maximum number of keypoints to keep.
    previous_results : np.ndarray
        Previous results from the model. Previous results are used to match keypoints between two frames.
    trigger : bool
        Trigger to set the reference frame.

    Output Message/s
    ----------------
    **Type**: dai.TrackedFeatures

    **Description**: TrackedFeatures message containing matched keypoints with the same ID.

    Error Handling
    --------------
    **ValueError**: If the original image size is not specified.
    **ValueError**: If the input image size is not specified.
    **ValueError**: If the maximum number of keypoints is not specified.
    **ValueError**: If the output layer containing features is not specified.
    **ValueError**: If the output layer containing keypoints is not specified.
    **ValueError**: If the output layer containing heatmaps is not specified.
    """

    def __init__(
        self,
        output_layer_feats: str = "feats",
        output_layer_keypoints: str = "keypoints",
        output_layer_heatmaps: str = "heatmaps",
        original_size: Tuple[float, float] = None,
        input_size: Tuple[float, float] = (640, 352),
        max_keypoints: int = 4096,
    ) -> None:
        """Initializes the XFeatParser node.

        @param output_layer_feats: Name of the output layer containing features.
        @type output_layer_feats: str
        @param output_layer_keypoints: Name of the output layer containing keypoints.
        @type output_layer_keypoints: str
        @param output_layer_heatmaps: Name of the output layer containing heatmaps.
        @type output_layer_heatmaps: str
        @param original_size: Original image size.
        @type original_size: Tuple[float, float]
        @param input_size: Input image size.
        @type input_size: Tuple[float, float]
        @param max_keypoints: Maximum number of keypoints to keep.
        @type max_keypoints: int
        """
        super().__init__(
            output_layer_feats,
            output_layer_keypoints,
            output_layer_heatmaps,
            original_size,
            input_size,
            max_keypoints,
        )

        self.previous_results = None
        self.trigger = False

        self._logger.debug(
            f"XFeatMonoParser initialized with output_layer_feats='{output_layer_feats}', output_layer_keypoints='{output_layer_keypoints}', output_layer_heatmaps='{output_layer_heatmaps}', original_size={original_size}, input_size={input_size}, max_keypoints={max_keypoints}"
        )

    def setTrigger(self) -> None:
        """Sets the trigger to set the reference frame."""
        self.trigger = True
        self._logger.debug(f"Trigger set to {self.trigger}")

    def run(self):
        self._logger.debug("XFeatMonoParser run started")
        self.validateParams()

        resize_rate_w = self.original_size[0] / self.input_size[0]
        resize_rate_h = self.original_size[1] / self.input_size[1]

        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            self._logger.debug(
                f"Processing input with layers: {output.getAllLayerNames()}"
            )
            feats, keypoints, heatmaps = self.extractTensors(output)

            result = detect_and_compute(
                feats,
                keypoints,
                heatmaps,
                resize_rate_w,
                resize_rate_h,
                self.input_size,
                self.max_keypoints,
            )

            if result is not None:
                result = result[0]
            else:
                matched_points = dai.TrackedFeatures()
                matched_points.setTimestamp(output.getTimestamp())
                matched_points.setSequenceNum(output.getSequenceNum())
                self._logger.debug(
                    "No keypoints found, sending TrackedFeatures message"
                )
                self.out.send(matched_points)
                self._logger.debug("TrackedFeatures message sent")
                continue

            if self.previous_results is not None:
                mkpts0, mkpts1 = match(self.previous_results, result)
                matched_points = create_tracked_features_message(mkpts0, mkpts1)
                matched_points.setTimestamp(output.getTimestamp())
                matched_points.setSequenceNum(output.getSequenceNum())
                self._logger.debug("Keypoints found, sending TrackedFeatures message")
                self.out.send(matched_points)
                self._logger.debug("TrackedFeatures message sent")
            else:
                matched_points = dai.TrackedFeatures()
                matched_points.setTimestamp(output.getTimestamp())
                matched_points.setSequenceNum(output.getSequenceNum())
                self._logger.debug(
                    "No previous results, sending TrackedFeatures message"
                )
                self.out.send(matched_points)
                self._logger.debug("TrackedFeatures message sent")

            if self.trigger:
                self.previous_results = result
                self.trigger = False


class XFeatStereoParser(XFeatBaseParser):
    """Parser class for parsing the output of the XFeat model. It can be used for parsing the output from two sources (e.g. two cameras - left and right).

    Attributes
    ----------
    reference_input : Node.Input
        Node's input. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    target_input : Node.Input
        Node's input. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    out : Node.Output
        Parser sends the processed network results to this output in a form of DepthAI message. It is a linking point from which the processed network results are retrieved.
    output_layer_feats : str
        Name of the output layer from which the features are extracted.
    output_layer_keypoints : str
        Name of the output layer from which the keypoints are extracted.
    output_layer_heatmaps : str
        Name of the output layer from which the heatmaps are extracted.
    original_size : Tuple[float, float]
        Original image size.
    input_size : Tuple[float, float]
        Input image size.
    max_keypoints : int
        Maximum number of keypoints to keep.

    Output Message/s
    ----------------
    **Type**: dai.TrackedFeatures

    **Description**: TrackedFeatures message containing matched keypoints with the same ID.

    Error Handling
    --------------
    **ValueError**: If the original image size is not specified.
    **ValueError**: If the input image size is not specified.
    **ValueError**: If the maximum number of keypoints is not specified.
    **ValueError**: If the output layer containing features is not specified.
    **ValueError**: If the output layer containing keypoints is not specified.
    **ValueError**: If the output layer containing heatmaps is not specified.
    """

    def __init__(
        self,
        output_layer_feats: str = "feats",
        output_layer_keypoints: str = "keypoints",
        output_layer_heatmaps: str = "heatmaps",
        original_size: Tuple[float, float] = None,
        input_size: Tuple[float, float] = (640, 352),
        max_keypoints: int = 4096,
    ) -> None:
        """Initializes the XFeatParser node.

        @param output_layer_feats: Name of the output layer containing features.
        @type output_layer_feats: str
        @param output_layer_keypoints: Name of the output layer containing keypoints.
        @type output_layer_keypoints: str
        @param output_layer_heatmaps: Name of the output layer containing heatmaps.
        @type output_layer_heatmaps: str
        @param original_size: Original image size.
        @type original_size: Tuple[float, float]
        @param input_size: Input image size.
        @type input_size: Tuple[float, float]
        @param max_keypoints: Maximum number of keypoints to keep.
        @type max_keypoints: int
        """
        super().__init__(
            output_layer_feats,
            output_layer_keypoints,
            output_layer_heatmaps,
            original_size,
            input_size,
            max_keypoints,
        )

        self._logger.debug(
            f"XFeatStereoParser initialized with output_layer_feats='{output_layer_feats}', output_layer_keypoints='{output_layer_keypoints}', output_layer_heatmaps='{output_layer_heatmaps}', original_size={original_size}, input_size={input_size}, max_keypoints={max_keypoints}"
        )

    def run(self):
        self._logger.debug("XFeatStereoParser run started")
        self.validateParams()

        resize_rate_w = self.original_size[0] / self.input_size[0]
        resize_rate_h = self.original_size[1] / self.input_size[1]

        while self.isRunning():
            try:
                reference_output: dai.NNData = self.reference_input.get()
                target_output: dai.NNData = self.target_input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            self._logger.debug(
                f"Processing reference input with layers: {reference_output.getAllLayerNames()}"
            )
            self._logger.debug(
                f"Processing target input with layers: {target_output.getAllLayerNames()}"
            )

            (
                reference_feats,
                reference_keypoints,
                reference_heatmaps,
            ) = self.extractTensors(reference_output)
            target_feats, target_keypoints, target_heatmaps = self.extractTensors(
                target_output
            )

            reference_result = detect_and_compute(
                reference_feats,
                reference_keypoints,
                reference_heatmaps,
                resize_rate_w,
                resize_rate_h,
                self.input_size,
                self.max_keypoints,
            )

            target_result = detect_and_compute(
                target_feats,
                target_keypoints,
                target_heatmaps,
                resize_rate_w,
                resize_rate_h,
                self.input_size,
                self.max_keypoints,
            )

            if reference_result is not None:
                reference_result = reference_result[0]
            else:
                matched_points = dai.TrackedFeatures()
                matched_points.setTimestamp(reference_output.getTimestamp())
                matched_points.setSequenceNum(reference_output.getSequenceNum())
                self._logger.debug(
                    "No reference keypoints found, sending TrackedFeatures message"
                )
                self.out.send(matched_points)
                self._logger.debug("TrackedFeatures message sent")
                continue

            if target_result is not None:
                target_result = target_result[0]
            else:
                matched_points = dai.TrackedFeatures()
                matched_points.setTimestamp(target_output.getTimestamp())
                matched_points.setSequenceNum(reference_output.getSequenceNum())
                self._logger.debug(
                    "No target keypoints found, sending TrackedFeatures message"
                )
                self.out.send(matched_points)
                self._logger.debug("TrackedFeatures message sent")
                continue

            mkpts0, mkpts1 = match(reference_result, target_result)
            matched_points = create_tracked_features_message(mkpts0, mkpts1)
            matched_points.setTimestamp(target_output.getTimestamp())
            matched_points.setSequenceNum(reference_output.getSequenceNum())

            self._logger.debug("Keypoints found, sending TrackedFeatures message")
            self.out.send(matched_points)
            self._logger.debug("TrackedFeatures message sent")



from typing import Any, Dict, List, Optional, Tuple

import depthai as dai
import numpy as np

from depthai_nodes.message.creators import create_keypoints_message
from depthai_nodes.node.parsers.keypoints import KeypointParser
from depthai_nodes.node.parsers.utils.superanimal import get_pose_prediction


class SuperAnimalParser(KeypointParser):
    """Parser class for parsing the output of the SuperAnimal landmark model.

    Attributes
    ----------
    output_layer_name: str
        Name of the output layer relevant to the parser.
    scale_factor : float
        Scale factor to divide the keypoints by.
    n_keypoints : int
        Number of keypoints.
    score_threshold : float
        Confidence score threshold for detected keypoints.
    label_names : List[str]
        Label names for the keypoints.
    edges : List[Tuple[int, int]]
        Keypoint connection pairs for visualizing the skeleton. Example: [(0,1), (1,2), (2,3), (3,0)] shows that keypoint 0 is connected to keypoint 1, keypoint 1 is connected to keypoint 2, etc.

    Output Message/s
    ----------------
    **Type**: Keypoints

    **Description**: Keypoints message containing detected keypoints that exceeds confidence threshold.
    """

    def __init__(
        self,
        output_layer_name: str = "",
        scale_factor: float = 256.0,
        n_keypoints: int = 39,
        score_threshold: float = 0.5,
        label_names: Optional[List[str]] = None,
        edges: Optional[List[Tuple[int, int]]] = None,
    ) -> None:
        """Initializes the parser node.

        @param output_layer_name: Name of the output layer relevant to the parser.
        @type output_layer_name: str
        @param n_keypoints: Number of keypoints.
        @type n_keypoints: int
        @param score_threshold: Confidence score threshold for detected keypoints.
        @type score_threshold: float
        @param scale_factor: Scale factor to divide the keypoints by.
        @type scale_factor: float
        @param label_names: Label names for the keypoints.
        @type label_names: Optional[List[str]]
        @param edges: Keypoint connection pairs for visualizing the skeleton. Example:
            [(0,1), (1,2), (2,3), (3,0)] shows that keypoint 0 is connected to keypoint
            1, keypoint 1 is connected to keypoint 2, etc.
        @type edges: Optional[List[Tuple[int, int]]]
        """
        super().__init__(
            output_layer_name,
            scale_factor=scale_factor,
            n_keypoints=n_keypoints,
            score_threshold=score_threshold,
            label_names=label_names,
            edges=edges,
        )
        self._logger.debug(
            f"SuperAnimalParser initialized with output_layer_name='{output_layer_name}', scale_factor={scale_factor}, n_keypoints={n_keypoints}, score_threshold={score_threshold}, label_names={label_names}, edges={edges}"
        )

    def build(
        self,
        head_config: Dict[str, Any],
    ) -> "SuperAnimalParser":
        """Configures the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: SuperAnimalParser
        """

        super().build(head_config)

        self._logger.debug("SuperAnimalParser built")

        return self

    def run(self):
        self._logger.debug("SuperAnimalParser run started")
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            layers = output.getAllLayerNames()
            self._logger.debug(f"Processing input with layers: {layers}")
            if len(layers) == 1 and self.output_layer_name == "":
                self.output_layer_name = layers[0]
            elif len(layers) != 1 and self.output_layer_name == "":
                raise ValueError(
                    f"Expected 1 output layer, got {len(layers)} layers. Please provide the output_layer_name."
                )
            heatmaps = output.getTensor(self.output_layer_name, dequantize=True).astype(
                np.float32
            )

            heatmaps_scale_factor = (
                self.scale_factor / heatmaps.shape[1],
                self.scale_factor / heatmaps.shape[2],
            )

            keypoints = get_pose_prediction(heatmaps, None, heatmaps_scale_factor)[0][0]
            scores = keypoints[:, 2]
            keypoints = keypoints[:, :2] / self.scale_factor

            msg = create_keypoints_message(
                keypoints,
                scores,
                self.score_threshold,
                label_names=self.label_names,
                edges=self.edges,
            )
            msg.setTimestamp(output.getTimestamp())
            msg.setTransformation(output.getTransformation())
            msg.setSequenceNum(output.getSequenceNum())

            self._logger.debug(
                f"Created keypoints message with {len(keypoints)} points"
            )

            self.out.send(msg)

            self._logger.debug("Keypoints message sent successfully")



from typing import Any, Dict

import depthai as dai
import numpy as np

from depthai_nodes.message.creators import create_detection_message
from depthai_nodes.node.parsers.detection import DetectionParser
from depthai_nodes.node.parsers.utils.ppdet import parse_paddle_detection_outputs


class PPTextDetectionParser(DetectionParser):
    """Parser class for parsing the output of the PaddlePaddle OCR text detection model.

    Attributes
    ----------
    output_layer_name: str
        Name of the output layer relevant to the parser.
    conf_threshold : float
        The threshold for bounding boxes.
    mask_threshold : float
        The threshold for the mask.
    max_det : int
        The maximum number of candidate bounding boxes.

    Output Message/s
    -------
    **Type**: dai.ImgDetections
    **Description**: ImgDetections message containing bounding boxes and the respective confidence scores of detected text.
    """

    def __init__(
        self,
        output_layer_name: str = "",
        conf_threshold: float = 0.5,
        mask_threshold: float = 0.25,
        max_det: int = 100,
    ) -> None:
        """Initializes the parser node.

        @param output_layer_name: Name of the output layer relevant to the parser.
        @type output_layer_name: str
        @param conf_threshold: The threshold for bounding boxes.
        @type conf_threshold: float
        @param mask_threshold: The threshold for the mask.
        @type mask_threshold: float
        @param max_det: The maximum number of candidate bounding boxes.
        @type max_det:
        """
        super().__init__(
            conf_threshold=conf_threshold,
            iou_threshold=0.5,
            max_det=max_det,
        )
        self.mask_threshold = mask_threshold
        self.output_layer_name = output_layer_name
        self._logger.debug(
            f"PPTextDetectionParser initialized with output_layer_name='{output_layer_name}', conf_threshold={conf_threshold}, mask_threshold={mask_threshold}, max_det={max_det}"
        )

    def setOutputLayerName(self, output_layer_name: str) -> None:
        """Sets the name of the output layer.

        @param output_layer_name: The name of the output layer.
        @type output_layer_name: str
        """
        if not isinstance(output_layer_name, str):
            raise ValueError("Output layer name must be a string.")
        self.output_layer_name = output_layer_name
        self._logger.debug(f"Output layer name set to '{self.output_layer_name}'")

    def setMaskThreshold(self, mask_threshold: float = 0.25) -> None:
        """Sets the mask threshold for creating the mask from model output
        probabilities.

        @param threshold: The threshold for the mask.
        @type threshold: float
        """
        if not isinstance(mask_threshold, float):
            raise ValueError("Mask threshold must be a float.")
        self.mask_threshold = mask_threshold
        self._logger.debug(f"Mask threshold set to {self.mask_threshold}")

    def build(self, head_config: Dict[str, Any]) -> "PPTextDetectionParser":
        """Configures the parser.

        @param config: The head configuration for the parser.
        @type config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: PPTextDetectionParser
        """

        super().build(head_config)
        self.mask_threshold = head_config.get("mask_threshold", self.mask_threshold)

        self._logger.debug(
            f"PPTextDetectionParser built with output_layer_name='{self.output_layer_name}', conf_threshold={self.conf_threshold}, mask_threshold={self.mask_threshold}, max_det={self.max_det}"
        )

        return self

    def run(self):
        self._logger.debug("PPTextDetectionParser run started")
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            layers = output.getAllLayerNames()
            self._logger.debug(f"Processing input with layers: {layers}")
            if len(layers) == 1 and self.output_layer_name == "":
                self.output_layer_name = layers[0]
            elif len(layers) != 1 and self.output_layer_name == "":
                raise ValueError(
                    f"Expected 1 output layer, got {len(layers)} layers. Please provide the output_layer_name."
                )

            predictions = np.array(
                output.getTensor(
                    self.output_layer_name,
                    dequantize=True,
                    storageOrder=dai.TensorInfo.StorageOrder.NCHW,
                )
            )

            _, _, height, width = predictions.shape

            bboxes, angles, scores = parse_paddle_detection_outputs(
                predictions,
                self.mask_threshold,
                self.conf_threshold,
                self.max_det,
                width=width,
                height=height,
            )
            message = create_detection_message(
                bboxes=bboxes, scores=scores, angles=angles
            )
            message.setTimestamp(output.getTimestamp())
            message.setTransformation(output.getTransformation())
            message.setSequenceNum(output.getSequenceNum())

            self._logger.debug(
                f"Created text detection message with {len(bboxes)} detections"
            )

            self.out.send(message)

            self._logger.debug("Text detection message sent successfully")



from tests.utils import HostNodeMock, PipelineMock, SyncMock, ThreadedHostNodeMock


def pytest_configure():
    import depthai as dai

    dai.Pipeline = PipelineMock
    dai.node.ThreadedHostNode = ThreadedHostNodeMock
    dai.node.HostNode = HostNodeMock
    dai.node.Sync = SyncMock



from typing import Any, Dict, List

import depthai as dai
import numpy as np

from depthai_nodes.message.creators import create_classification_sequence_message
from depthai_nodes.node.parsers.utils import softmax

from .classification import ClassificationParser


class ClassificationSequenceParser(ClassificationParser):
    """Postprocessing logic for a classification sequence model. The model predicts the
    classes multiple times and returns a list of predicted classes, where each item
    corresponds to the relative step in the sequence. In addition to time series
    classification, this parser can also be used for text recognition models where words
    can be interpreted as a sequence of characters (classes).

    Attributes
    ----------
    output_layer_name: str
        Name of the output layer relevant to the parser.
    classes: List[str]
        List of available classes for the model.
    is_softmax: bool
        If False, the scores are converted to probabilities using softmax function.
    ignored_indexes: List[int]
        List of indexes to ignore during classification generation (e.g., background class, blank space).
    remove_duplicates: bool
        If True, removes consecutive duplicates from the sequence.
    concatenate_classes: bool
        If True, concatenates consecutive words based on the predicted spaces.

    Output Message/s
    ----------------
    **Type**: Classifications(dai.Buffer)

    **Description**:
        An object with attributes `classes` and `scores`. `classes` is a list containing the predicted classes. `scores` is a list of corresponding probability scores.
    """

    def __init__(
        self,
        output_layer_name: str = "",
        classes: List[str] = None,
        is_softmax: bool = True,
        ignored_indexes: List[int] = None,
        remove_duplicates: bool = False,
        concatenate_classes: bool = False,
    ) -> None:
        """Initializes the parser node.

        @param output_layer_name: Name of the output layer relevant to the parser.
        @type output_layer_name: str
        @param classes: List of available classes for the model.
        @type classes: List[str]
        @param ignored_indexes: List of indexes to ignore during classification
            generation (e.g., background class, blank space).
        @type ignored_indexes: List[int]
        @param is_softmax: If False, the scores are converted to probabilities using
            softmax function.
        @type is_softmax: bool
        @param remove_duplicates: If True, removes consecutive duplicates from the
            sequence.
        @type remove_duplicates: bool
        @param concatenate_classes: If True, concatenates consecutive words based on the
            predicted spaces.
        @type concatenate_classes: bool
        """
        super().__init__(
            output_layer_name=output_layer_name, classes=classes, is_softmax=is_softmax
        )

        self.ignored_indexes = ignored_indexes if ignored_indexes is not None else []
        self.remove_duplicates = remove_duplicates
        self.concatenate_classes = concatenate_classes
        self._logger.debug(
            f"ClassificationSequenceParser initialized with output_layer_name='{output_layer_name}', classes={classes}, is_softmax={is_softmax}, ignored_indexes={ignored_indexes}, remove_duplicates={remove_duplicates}, concatenate_classes={concatenate_classes}"
        )

    def setRemoveDuplicates(self, remove_duplicates: bool) -> None:
        """Sets the remove_duplicates flag for the classification sequence model.

        @param remove_duplicates: If True, removes consecutive duplicates from the
            sequence.
        @type remove_duplicates: bool
        """
        if not isinstance(remove_duplicates, bool):
            raise ValueError("remove_duplicates must be a boolean.")
        self.remove_duplicates = remove_duplicates
        self._logger.debug(f"Remove duplicates set to {remove_duplicates}")

    def setIgnoredIndexes(self, ignored_indexes: List[int]) -> None:
        """Sets the ignored_indexes for the classification sequence model.

        @param ignored_indexes: A list of indexes to ignore during classification
            generation.
        @type ignored_indexes: List[int]
        """
        if not isinstance(ignored_indexes, list):
            raise ValueError("Ignored indexes must be a list.")
        if not all(isinstance(index, int) for index in ignored_indexes):
            raise ValueError("All ignored indexes must be integers.")
        self.ignored_indexes = ignored_indexes
        self._logger.debug(f"Ignored indexes set to {ignored_indexes}")

    def setConcatenateClasses(self, concatenate_classes: bool) -> None:
        """Sets the concatenate_classes flag for the classification sequence model.

        @param concatenate_classes: If True, concatenates consecutive classes into a
            single string. Used mostly for text processing.
        @type concatenate_classes: bool
        """
        if not isinstance(concatenate_classes, bool):
            raise ValueError("concatenate_classes must be a boolean.")
        self.concatenate_classes = concatenate_classes
        self._logger.debug(f"Concatenate classes set to {concatenate_classes}")

    def build(self, head_config: Dict[str, Any]) -> "ClassificationSequenceParser":
        """Configures the parser.

        @param head_config: The head configuration for the parser. The required keys are `classes`, `n_classes`, and `is_softmax`.
        In addition to these, there are three optional keys that are mostly used for text processing: `ignored_indexes`, `remove_duplicates` and `concatenate_classes`.
        @type head_config: Dict[str, Any]

        @return: Returns the instantiated parser with the correct configuration.
        @rtype: ClassificationParser
        """
        super().build(head_config)
        self.ignored_indexes = head_config.get("ignored_indexes", [])
        self.remove_duplicates = head_config.get(
            "remove_duplicates", self.remove_duplicates
        )
        self.concatenate_classes = head_config.get(
            "concatenate_classes", self.concatenate_classes
        )

        self._logger.debug(
            f"ClassificationSequenceParser built with ignored_indexes={self.ignored_indexes}, remove_duplicates={self.remove_duplicates}, concatenate_classes={self.concatenate_classes}"
        )

        return self

    def run(self):
        self._logger.debug("ClassificationSequenceParser run started")
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()

            except dai.MessageQueue.QueueException:
                break

            layers = output.getAllLayerNames()
            self._logger.debug(f"Processing input with layers: {layers}")
            if len(layers) == 1 and self.output_layer_name == "":
                self.output_layer_name = layers[0]
            elif len(layers) != 1 and self.output_layer_name == "":
                raise ValueError(
                    f"Expected 1 output layer, got {len(layers)} layers. Please provide the output_layer_name."
                )

            if self.n_classes == 0:
                raise ValueError("Classes must be provided for classification.")

            scores = output.getTensor(self.output_layer_name, dequantize=True).astype(
                np.float32
            )

            if len(scores.shape) != 3 and len(scores.shape) != 2:
                raise ValueError(
                    f"Scores should be a 3D or 2D array, got shape {scores.shape}."
                )

            if len(scores.shape) == 3:
                if scores.shape[0] == 1:
                    scores = scores[0]
                elif scores.shape[2] == 1:
                    scores = scores[:, :, 0]
                else:
                    raise ValueError(
                        "Scores should be a 3D array of shape (1, sequence_length, n_classes) or (sequence_length, n_classes, 1)."
                    )

            if not self.is_softmax:
                scores = softmax(scores, axis=1, keep_dims=True)

            msg = create_classification_sequence_message(
                classes=self.classes,
                scores=scores,
                remove_duplicates=self.remove_duplicates,
                ignored_indexes=self.ignored_indexes,
                concatenate_classes=self.concatenate_classes,
            )
            msg.setTransformation(output.getTransformation())
            msg.setTimestamp(output.getTimestamp())
            msg.setSequenceNum(output.getSequenceNum())

            self._logger.debug(f"Created message with {len(msg.classes)} classes")

            self.out.send(msg)

            self._logger.debug("Classification message sent successfully")



import copy
from typing import List

import depthai as dai

from depthai_nodes import PRIMARY_COLOR

from .utils import (
    copy_message,
)


class Line(dai.Buffer):
    """Line class for storing a line.

    Attributes
    ----------
    start_point : dai.Point2f
        Start point of the line with x and y coordinate.
    end_point : dai.Point2f
        End point of the line with x and y coordinate.
    confidence : float
        Confidence of the line.
    """

    def __init__(self):
        """Initializes the Line object."""
        super().__init__()
        self._start_point: dai.Point2f = None
        self._end_point: dai.Point2f = None
        self._confidence: float = None

    def copy(self):
        """Creates a new instance of the Line class and copies the attributes.

        @return: A new instance of the Line class.
        @rtype: Line
        """
        new_obj = Line()
        new_obj.start_point = copy_message(self.start_point)
        new_obj.end_point = copy_message(self.end_point)
        new_obj.confidence = copy.deepcopy(self.confidence)
        return new_obj

    @property
    def start_point(self) -> dai.Point2f:
        """Returns the start point of the line.

        @return: Start point of the line.
        @rtype: dai.Point2f
        """
        return self._start_point

    @start_point.setter
    def start_point(self, value: dai.Point2f):
        """Sets the start point of the line.

        @param value: Start point of the line.
        @type value: dai.Point2f
        @raise TypeError: If value is not of type dai.Point2f.
        """
        if not isinstance(value, dai.Point2f):
            raise TypeError(
                f"Start Point must be of type Point2f, instead got {type(value)}."
            )
        self._start_point = value

    @property
    def end_point(self) -> dai.Point2f:
        """Returns the end point of the line.

        @return: End point of the line.
        @rtype: dai.Point2f
        """
        return self._end_point

    @end_point.setter
    def end_point(self, value: dai.Point2f):
        """Sets the end point of the line.

        @param value: End point of the line.
        @type value: dai.Point2f
        @raise TypeError: If value is not of type dai.Point2f.
        """
        if not isinstance(value, dai.Point2f):
            raise TypeError(
                f"End Point must be of type Point2f, instead got {type(value)}."
            )
        self._end_point = value

    @property
    def confidence(self) -> float:
        """Returns the confidence of the line.

        @return: Confidence of the line.
        @rtype: float
        """
        return self._confidence

    @confidence.setter
    def confidence(self, value: float):
        """Sets the confidence of the line.

        @param value: Confidence of the line.
        @type value: float
        @raise TypeError: If value is not of type float.
        """
        if not isinstance(value, float):
            raise TypeError(
                f"Confidence must be of type float, instead got {type(value)}."
            )
        self._confidence = value


class Lines(dai.Buffer):
    """Lines class for storing lines.

    Attributes
    ----------
    lines : List[Line]
        List of detected lines.
    transformation : dai.ImgTransformation
        Image transformation object.
    """

    def __init__(self):
        """Initializes the Lines object."""
        super().__init__()
        self._lines: List[Line] = []
        self._transformation: dai.ImgTransformation = None

    def copy(self):
        """Creates a new instance of the Lines class and copies the attributes.

        @return: A new instance of the Lines class.
        @rtype: Lines
        """
        new_obj = Lines()
        new_obj.lines = [line.copy() for line in self.lines]
        new_obj.setSequenceNum(self.getSequenceNum())
        new_obj.setTimestamp(self.getTimestamp())
        new_obj.setTimestampDevice(self.getTimestampDevice())
        new_obj.setTransformation(self.transformation)
        return new_obj

    @property
    def lines(self) -> List[Line]:
        """Returns the lines.

        @return: List of lines.
        @rtype: List[Line]
        """
        return self._lines

    @lines.setter
    def lines(self, value: List[Line]):
        """Sets the lines.

        @param value: List of lines.
        @type value: List[Line]
        @raise TypeError: If value is not a list.
        @raise TypeError: If each element is not of type Line.
        """
        if not isinstance(value, List):
            raise TypeError(f"lines must be a list, instead got {type(value)}.")
        if not all(isinstance(item, Line) for item in value):
            raise ValueError("Lines must be a list of Line objects.")
        self._lines = value

    @property
    def transformation(self) -> dai.ImgTransformation:
        """Returns the Image Transformation object.

        @return: The Image Transformation object.
        @rtype: dai.ImgTransformation
        """
        return self._transformation

    @transformation.setter
    def transformation(self, value: dai.ImgTransformation):
        """Sets the Image Transformation object.

        @param value: The Image Transformation object.
        @type value: dai.ImgTransformation
        @raise TypeError: If value is not a dai.ImgTransformation object.
        """

        if value is not None:
            if not isinstance(value, dai.ImgTransformation):
                raise TypeError(
                    f"Transformation must be a dai.ImgTransformation object, instead got {type(value)}."
                )
        self._transformation = value

    def setTransformation(self, transformation: dai.ImgTransformation):
        """Sets the Image Transformation object.

        @param transformation: The Image Transformation object.
        @type transformation: dai.ImgTransformation
        @raise TypeError: If value is not a dai.ImgTransformation object.
        """
        self.transformation = transformation

    def getVisualizationMessage(self) -> dai.ImgAnnotations:
        """Returns default visualization message for lines.

        The message adds lines to the image.
        """
        img_annotation = dai.ImgAnnotations()
        annotation = dai.ImgAnnotation()

        for line in self.lines:
            pointsAnnotation = dai.PointsAnnotation()
            pointsAnnotation.type = dai.PointsAnnotationType.LINE_STRIP
            pointsAnnotation.points = dai.VectorPoint2f(
                [line.start_point, line.end_point]
            )
            pointsAnnotation.outlineColor = PRIMARY_COLOR
            pointsAnnotation.thickness = 2.0
            annotation.points.append(pointsAnnotation)

        img_annotation.annotations.append(annotation)
        img_annotation.setTimestamp(self.getTimestamp())
        return img_annotation



import numpy as np


def normalize_keypoints(keypoints: np.ndarray, height: int, width: int) -> np.ndarray:
    """Normalize keypoint coordinates to (0, 1).

    Parameters:
    @param keypoints: A numpy array of shape (N, 2) or (N, K, 2) where N is the number of keypoint sets and K is the number of keypoint in each set.
    @type np.ndarray
    @param height: The height of the image.
    @type height: int
    @param width: The width of the image.
    @type width: int

    Returns:
    np.ndarray: A numpy array of shape (N, 2) containing the normalized keypoints.
    """
    keypoints = keypoints.astype(np.float32)
    if not isinstance(keypoints, np.ndarray):
        raise TypeError("Keypoints must be a numpy array.")

    if len(keypoints.shape) != 2 and len(keypoints.shape) != 3:
        raise ValueError(
            f"Keypoints must be of shape (N, 2) or (N, K, 2). Got {keypoints.shape}."
        )

    if keypoints.shape[1] != 2 and len(keypoints.shape) == 2:
        raise ValueError(
            "Keypoints must be of shape (N, 2). Other options are currently not supported"
        )
    elif len(keypoints.shape) == 3:
        if keypoints.shape[2] != 2:
            raise ValueError(
                "Keypoints must be of shape (N, K, 2). Other options are currently not supported"
            )
    keypoints[:, 0] = keypoints[:, 0] / width
    keypoints[:, 1] = keypoints[:, 1] / height

    return keypoints



from typing import Optional

import numpy as np


def softmax(
    x: np.ndarray, axis: Optional[int] = None, keep_dims: bool = False
) -> np.ndarray:
    """Compute the softmax of an array. The softmax function is defined as: softmax(x) =
    exp(x) / sum(exp(x))

    @param x: The input array.
    @type x: np.ndarray
    @param axis: Axis or axes along which a sum is performed. The default, axis=None,
        will sum all of the elements of the input array. If axis is negative it counts
        from the last to the first axis.
    @type axis: int
    @param keep_dims: If this is set to True, the axes which are reduced are left in the
        result as dimensions with size one. With this option, the result will broadcast
        correctly against the input array.
    @type keep_dims: bool
    @return: The softmax of the input array.
    @rtype: np.ndarray
    """
    ex = np.exp(x)
    return ex / np.sum(ex, axis=axis, keepdims=keep_dims)


def sigmoid(x: np.ndarray) -> np.ndarray:
    """Sigmoid function.

    @param x: Input tensor.
    @type x: np.ndarray
    @return: A result tensor after applying a sigmoid function on the given input.
    @rtype: np.ndarray
    """
    return 1 / (1 + np.exp(-x))



from typing import Optional, Tuple

import cv2
import numpy as np

from depthai_nodes.node.parsers.utils import (
    corners_to_rotated_bbox,
)


def _get_mini_boxes(contour: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Internal function to get the minimum bounding box of a contour.

    @param contour: The contour to get the minimum bounding box of the text.
    @type contour: np.ndarray
    @return: The minimum rotated bounding box defined as [x_center, y_center, width,
        height, angle] and the corners of the box.
    @rtype: Tuple[np.ndarray, np.ndarray]
    """
    bounding_box = cv2.minAreaRect(contour)
    points = sorted(list(cv2.boxPoints(bounding_box)), key=lambda x: x[0])

    index_1, index_2, index_3, index_4 = 0, 1, 2, 3
    if points[1][1] > points[0][1]:
        index_1 = 0
        index_4 = 1
    else:
        index_1 = 1
        index_4 = 0

    if points[3][1] > points[2][1]:
        index_2 = 2
        index_3 = 3
    else:
        index_2 = 3
        index_3 = 2

    corners = [points[index_1], points[index_2], points[index_3], points[index_4]]

    bbox = corners_to_rotated_bbox(np.array(corners))
    return bbox, np.array(corners)


def _dilate_box(corners: np.ndarray, pixels: int) -> np.ndarray:
    """Internal function to dilate the bounding box area by a specified number of
    pixels.

    @param corners: The corners of the bounding box.
    @type corners: np.ndarray
    @param pixels: The number of pixels to dilate the bounding box area by.
    @type pixels: int
    @return: The dilated bounding box corners.
    @rtype: np.ndarray
    """
    corners[0] = corners[0] - pixels
    corners[1][0] = corners[1][0] + pixels
    corners[1][1] = corners[1][1] - pixels
    corners[2] = corners[2] + pixels
    corners[3][0] = corners[3][0] - pixels
    corners[3][1] = corners[3][1] + pixels

    return corners


def _box_score(predictions: np.ndarray, _corners: np.ndarray) -> float:
    """Internal function to calculate the score of a bounding box based on the mean
    pixel values within the box area.

    @params predictions: The predictions from the model.
    @type predictions: np.ndarray
    @params _corners: The corners of the bounding box.
    @type _corners: np.ndarray
    @return: The score of the bounding box.
    @rtype: float
    """
    h, w = predictions.shape[:2]
    corners = _dilate_box(_corners, -2)
    mask = np.zeros((h, w), dtype=np.uint8)
    cv2.fillPoly(mask, [corners.reshape(-1, 1, 2).astype(np.int32)], 1)

    return float(np.sum(predictions * mask) / (np.sum(mask) + 1e-5))


def _unclip(
    box: np.ndarray,
    unclip_ratio: float = 2,
) -> np.ndarray:
    """Internal function to dilate the bounding box area by a specified ratio.

    @param box: The rotated bounding box.
    @type box: np.ndarray
    @param unclip_ratio: The ratio to dilate the bounding box area by.
    @type unclip_ratio: float = 3
    @return: The dilated bounding box corners.
    @rtype: np.ndarray
    """

    box[2] = box[2] * np.sqrt(unclip_ratio)
    box[3] = box[3] * np.sqrt(unclip_ratio)

    return box


def parse_paddle_detection_outputs(
    predictions: np.ndarray,
    mask_threshold: float = 0.25,
    bbox_threshold: float = 0.5,
    max_detections: int = 100,
    width: Optional[int] = None,
    height: Optional[int] = None,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Parse the output of a PaddlePaddle Text Detection model from a mask of text
    probabilities into rotated bounding boxes with additional corners saved as
    keypoints.

    @param predictions: The output of a PaddlePaddle Text Detection model.
    @type predictions: np.ndarray
    @param mask_threshold: The threshold for the mask.
    @type mask_threshold: float
    @param bbox_threshold: The threshold for bounding boxes.
    @type bbox_threshold: float
    @param max_detections: The maximum number of candidate bounding boxes.
    @type max_detections: int
    @return: A touple containing the rotated bounding boxes, corners and scores.
    @rtype: Touple[np.ndarray, np.ndarray, np.ndarray]
    """

    if len(predictions.shape) == 4:
        if predictions.shape[0] == 1 and predictions.shape[1] == 1:
            predictions = predictions[0, 0]
        elif predictions.shape[0] == 1 and predictions.shape[3] == 1:
            predictions = predictions[0, :, :, 0]
        else:
            raise ValueError(
                f"Predictions should be either (1, 1, H, W) or (1, H, W, 1), got {predictions.shape}."
            )
    else:
        raise ValueError(
            f"Predictions should be 4D array of shape (1, 1, H, W) or (1, H, W, 1), got {predictions.shape}."
        )

    mask = predictions > mask_threshold
    kernel = np.ones((5, 5), np.uint8)
    mask = cv2.dilate(mask.astype(np.uint8), kernel, iterations=1)

    outs = cv2.findContours(
        (mask * 255).astype(np.uint8), cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE
    )

    if len(outs) == 3:
        _, contours, _ = outs[0], outs[1], outs[2]
    elif len(outs) == 2:
        contours, _ = outs[0], outs[1]
    contours = sorted(contours, key=cv2.contourArea, reverse=True)

    num_contours = min(len(contours), max_detections)

    boxes = []
    scores = []
    angles = []
    for contour in contours[:num_contours]:
        box, corners = _get_mini_boxes(contour)
        if min(box[2], box[3]) < 8:
            continue

        score = _box_score(predictions, corners.reshape(-1, 2))
        if score < bbox_threshold:
            continue

        box = _unclip(box)
        boxes.append(box[:4])
        scores.append(score)
        angles.append(box[4])

    boxes = np.array(boxes)
    if boxes.size > 0:
        boxes[:, 0] /= width
        boxes[:, 1] /= height
        boxes[:, 2] /= width
        boxes[:, 3] /= height

    boxes = np.clip(boxes, 0.0, 1.0)
    angles = np.array(angles)
    angles = np.round(angles, 0)

    return boxes, angles, np.array(scores)



import pytest

from depthai_nodes import Prediction, Predictions
from depthai_nodes.message.creators import create_regression_message


def test_valid_input():
    predictions = [0.1, 0.2, 0.3]
    message = create_regression_message(predictions)

    assert isinstance(message, Predictions)
    assert len(message.predictions) == 3
    assert all(isinstance(pred, Prediction) for pred in message.predictions)
    assert message.predictions[0].prediction == 0.1
    assert message.predictions[1].prediction == 0.2
    assert message.predictions[2].prediction == 0.3


def test_empty_list():
    predictions = []
    message = create_regression_message(predictions)

    assert isinstance(message, Predictions)
    assert len(message.predictions) == 0


def test_invalid_type():
    with pytest.raises(ValueError):
        create_regression_message("not a list")


def test_invalid_prediction_type():
    predictions = [0.1, "not a float", 0.3]
    with pytest.raises(ValueError):
        create_regression_message(predictions)



from pathlib import Path
from typing import Dict, Optional, Type, TypeVar, Union, overload

import depthai as dai

from depthai_nodes.logging import get_logger
from depthai_nodes.node.parser_generator import ParserGenerator
from depthai_nodes.node.parsers import BaseParser

TParser = TypeVar("TParser", bound=Union[BaseParser, dai.DeviceNode])


class ParsingNeuralNetwork(dai.node.ThreadedHostNode):
    Propeties = dai.node.NeuralNetwork.Properties
    """Node that wraps the NeuralNetwork node and adds parsing capabilities. A
    NeuralNetwork node is created with it's appropriate parser nodes for each model
    head. Parser nodes are chosen based on the supplied NNArchive.

    Attributes
    ----------
    input : Node.Input
        Neural network input.
    inputs : Node.InputMap
        Neural network inputs.
    out : Node.Output
        Neural network output. Can be used only when there is exactly one model head. Otherwise, getOutput method must be used.
    outputs: Node.Output
        Neural network output having dai.MessageGroup as a payload which contains outputs of all model heads and can be accessed as a dictionary with str(model head index) as a key. Can be used only when there is at least two model heads. Otherwise, out property must be used.
    passthrough : Node.Output
        Neural network passthrough.
    passthroughs : Node.OutputMap
        Neural network passthroughs.
    """

    def __init__(self, *args, **kwargs) -> None:
        """Initializes the ParsingNeuralNetwork node. NeuralNetwork node is created in
        the pipeline.

        @param args: Arguments to be passed to the ThreadedHostNode class.
        @param kwargs: Keyword arguments to be passed to the ThreadedHostNode class.
        """
        super().__init__(*args, **kwargs)
        self._pipeline = self.getParentPipeline()
        self._nn = self._pipeline.create(dai.node.NeuralNetwork)
        self._parsers: Dict[int, BaseParser] = {}
        self._internal_sync: Optional[dai.node.Sync] = None
        self._logger = get_logger(__name__)
        self._logger.debug("ParsingNeuralNetwork initialized")

    @property
    def input(self) -> dai.Node.Input:
        return self._nn.input

    @property
    def inputs(self) -> dai.Node.InputMap:
        return self._nn.inputs

    @property
    def out(self) -> dai.Node.Output:
        self._checkNNArchive()
        if len(self._parsers) != 1:
            raise RuntimeError(
                f"Property out is only available when there is exactly one model head. \
                               The model has {self._getModelHeadsLen()} heads. Use {self.getOutput.__name__} method or {self.outputs.__name__} property instead."
            )
        return list(self._parsers.values())[0].out

    @property
    def outputs(self) -> dai.Node.Output:
        """Neural network output having dai.MessageGroup as a payload which contains
        outputs of all model heads and can be accessed as a dictionary with str(model
        head index) as a key.

        Can be used only when there are at least two model heads. Otherwise, out
        property must be used.
        """
        if self._internal_sync is None:
            raise RuntimeError(
                f"ParsingNeuralNetwork node must have at least two model heads to use sync node and outputs property. The model has {self._getModelHeadsLen()} heads."
            )
        return self._internal_sync.out

    @property
    def passthrough(self) -> dai.Node.Output:
        return self._nn.passthrough

    @property
    def passthroughs(self) -> dai.Node.OutputMap:
        return self._nn.passthroughs

    def getNumInferenceThreads(self) -> int:
        """Returns number of inference threads of the NeuralNetwork node."""
        return self._nn.getNumInferenceThreads()

    @overload
    def getParser(self, index: int = 0) -> Union[BaseParser, dai.DeviceNode]:
        """Returns the parser node for the given model head index.

        If index is not provided, the first parser node is returned by default.

        @param index: Index of the model head parser to return.
        @type index: int
        @param parser_type: Expected type of the parser. Used for type checking and
            casting.
        @type parser_type: Type[TParser]
        @return: Parser node for the specified model head index, cast to the requested
            parser_type.
        @rtype: TParser
        @raise KeyError: If parser with given index doesn't exist.
        @raise TypeError: If parser is not of the requested type.
        """
        ...

    @overload
    def getParser(self, parser_type: Type[TParser], index: int = 0) -> TParser:
        """Returns the parser node for the given model head index.

        If index is not provided, the first parser node is returned by default.

        @param index: Index of the model head parser to return.
        @type index: int
        @param parser_type: Expected type of the parser. Used for type checking and
            casting.
        @type parser_type: Type[TParser]
        @return: Parser node for the specified model head index, cast to the requested
            parser_type.
        @rtype: TParser
        @raise KeyError: If parser with given index doesn't exist.
        @raise TypeError: If parser is not of the requested type.
        """
        ...

    def getParser(self, *args, **kwargs) -> Union[BaseParser, dai.DeviceNode]:
        index = 0
        parser_type = None

        # Parse arguments based on the overload patterns
        if len(args) == 1:
            if isinstance(args[0], type):  # Case: getParser(parser_type)
                parser_type = args[0]
            else:  # Case: getParser(index)
                index = args[0]
        elif len(args) == 2:  # Case: getParser(parser_type, index)
            parser_type = args[0]
            index = args[1]

        # Handle kwargs
        if "index" in kwargs:
            index = kwargs["index"]
        if "parser_type" in kwargs:
            parser_type = kwargs["parser_type"]

        if index not in self._parsers:
            raise KeyError(
                f"Parser with ID {index} not found. Available parser IDs: {list(self._parsers.keys())}"
            )
        parser = self._parsers[index]

        if parser_type:
            if not isinstance(parser, parser_type):
                raise TypeError(
                    f"Parser with ID {index} is of type: {type(parser)}. Requested type: {parser_type}"
                )
        return parser

    def getOutput(self, head: int) -> dai.Node.Output:
        """Obtains output of a parser for specified NeuralNetwork model head."""
        if head not in self._parsers:
            raise KeyError(
                f"Head {head} is not available. Available heads for the model {self._getModelName()} are {list(self._parsers.keys())}."
            )
        return self._parsers[head].out

    def setBackend(self, setBackend: str) -> None:
        """Sets the backend of the NeuralNetwork node."""
        self._nn.setBackend(setBackend)

    def setBackendProperties(self, setBackendProperties: Dict[str, str]) -> None:
        """Sets the backend properties of the NeuralNetwork node."""
        self._nn.setBackendProperties(setBackendProperties)

    def setBlob(self, blob: Union[Path, dai.OpenVINO.Blob]) -> None:
        """Sets the blob of the NeuralNetwork node."""
        self._nn.setBlob(blob)

    def setBlobPath(self, path: Path) -> None:
        """Sets the blob path of the NeuralNetwork node."""
        self._nn.setBlobPath(path)

    def setFromModelZoo(
        self, description: dai.NNModelDescription, useCached: bool
    ) -> None:
        """Sets the model from the model zoo of the NeuralNetwork node."""
        self._nn.setFromModelZoo(description, useCached)

    def setModelPath(self, modelPath: Path) -> None:
        """Sets the model path of the NeuralNetwork node."""
        self._nn.setModelPath(modelPath)

    def setNNArchive(
        self, nnArchive: dai.NNArchive, numShaves: Optional[int] = None
    ) -> None:
        """Sets the NNArchive of the ParsingNeuralNetwork node.

        Updates the NeuralNetwork node and parser nodes.

        @param nnArchive: Neural network archive containing the model and its
            configuration.
        @type nnArchive: dai.NNArchive
        @param numShaves: Optional number of shaves to allocate for the neural network.
            If not provided, uses default allocation.
        @type numShaves: Optional[int]
        """
        self._nn_archive = nnArchive
        if numShaves:
            self._nn.setNNArchive(nnArchive, numShaves)  # type: ignore
        else:
            self._nn.setNNArchive(nnArchive)
        self._updateParsers(nnArchive)

    def setNumInferenceThreads(self, numThreads: int) -> None:
        """Sets the number of inference threads of the NeuralNetwork node."""
        self._nn.setNumInferenceThreads(numThreads)

    def setNumNCEPerInferenceThread(self, numNCEPerThread: int) -> None:
        """Sets the number of NCE per inference thread of the NeuralNetwork node."""
        self._nn.setNumNCEPerInferenceThread(numNCEPerThread)

    def setNumPoolFrames(self, numFrames: int) -> None:
        """Sets the number of pool frames of the NeuralNetwork node."""
        self._nn.setNumPoolFrames(numFrames)

    def setNumShavesPerInferenceThread(self, numShavesPerInferenceThread: int) -> None:
        """Sets the number of shaves per inference thread of the NeuralNetwork node."""
        self._nn.setNumShavesPerInferenceThread(numShavesPerInferenceThread)

    def build(
        self,
        input: Union[dai.Node.Output, dai.node.Camera],
        nn_source: Union[dai.NNModelDescription, dai.NNArchive, str],
        fps: Optional[float] = None,
    ) -> "ParsingNeuralNetwork":
        """Builds the underlying NeuralNetwork node and creates parser nodes for each
        model head.

        @param input: Node's input. It is a linking point to which the NeuralNetwork is
            linked. It accepts the output of a Camera node.
        @type input: Node.Input

        @param nn_source: NNModelDescription object containing the HubAI model descriptors, NNArchive object of the model, or HubAI model slug in form of <model_slug>:<model_version_slug> or <model_slug>:<model_version_slug>:<model_instance_hash>.
        @type nn_source: Union[dai.NNModelDescription, dai.NNArchive, str]
        @param fps: FPS limit for the model runtime.
        @type fps: int
        @return: Returns the ParsingNeuralNetwork object.
        @rtype: ParsingNeuralNetwork
        @raise ValueError: If the nn_source is not a NNModelDescription or NNArchive
            object.
        """

        platform = self.getParentPipeline().getDefaultDevice().getPlatformAsString()

        if isinstance(nn_source, str):
            nn_source = dai.NNModelDescription(nn_source)
        if isinstance(nn_source, (dai.NNModelDescription, str)):
            if not nn_source.platform:
                nn_source.platform = platform
            self._nn_archive = dai.NNArchive(dai.getModelFromZoo(nn_source))
        elif isinstance(nn_source, dai.NNArchive):
            self._nn_archive = nn_source
        else:
            raise ValueError(
                "nn_source must be either a NNModelDescription, NNArchive, or a string representing HubAI model slug."
            )

        kwargs = {"fps": fps} if fps else {}
        self._nn.build(input, self._nn_archive, **kwargs)

        self._updateParsers(self._nn_archive)

        if len(self._parsers) > 1:
            self._createSyncNode()

        self._logger.debug(
            f"ParsingNeuralNetwork built with fps={fps}, type_of_nn_source={type(nn_source).__name__}, parsers={len(self._parsers)}"
        )

        return self

    def run(self) -> None:
        """Methods inherited from ThreadedHostNode.

        Method runs with start of the pipeline.
        """
        self._checkNNArchive()

    def cleanup(self):
        """Cleans up the ParsingNeuralNetwork node and removes all nodes created by
        ParsingNeuralNetwork from the pipeline.

        Must be called before removing the node from the pipeline.
        """
        self._pipeline.remove(self._nn)
        self._nn = None

        self._removeOldParserNodes()
        self._internal_sync = None
        self._parsers = {}

    def _updateParsers(self, nnArchive: dai.NNArchive) -> None:
        self._removeOldParserNodes()
        self._parsers = self._getParserNodes(nnArchive)

    def _removeOldParserNodes(self) -> None:
        for parser in self._parsers.values():
            self._pipeline.remove(parser)
        if self._internal_sync is not None:
            self._pipeline.remove(self._internal_sync)

    def _getParserNodes(self, nnArchive: dai.NNArchive) -> Dict[int, BaseParser]:
        parser_generator = self._pipeline.create(ParserGenerator)
        parsers = self._generateParsers(parser_generator, nnArchive)
        for parser in parsers.values():
            self._nn.out.link(
                parser.input
            )  # TODO: once NN node has output dictionary, link to the correct output
        self._pipeline.remove(parser_generator)
        return parsers

    def _generateParsers(
        self, parserGenerator: ParserGenerator, nnArchive: dai.NNArchive
    ) -> Dict[int, BaseParser]:
        return parserGenerator.build(nnArchive)

    def _getModelHeadsLen(self):
        heads = self._getModelHeads()
        if not heads:
            return 0
        return len(heads)

    def _getModelHeads(self):
        return self._getConfig().model.heads

    def _checkNNArchive(self) -> None:
        if self._nn_archive is None:
            raise RuntimeError(
                f"NNArchive is not set. Use {self.setNNArchive.__name__} or {self.build.__name__} method to set it."
            )

    def _getModelName(self) -> str:
        return self._getConfig().model.metadata.name

    def _getConfig(self):
        return self._nn_archive.getConfig()

    def _createSyncNode(self):
        if self._internal_sync is not None:
            raise RuntimeError(
                "Sync node already exists. Remove it before creating a new one."
            )
        if len(list(self._parsers.values())) <= 1:
            raise RuntimeError(
                "ParsingNeuralNetwork node must have at least two model heads to use sync node."
            )
        self._internal_sync = self._pipeline.create(dai.node.Sync)
        self._internal_sync.setRunOnHost(True)
        outputs = [parser.out for parser in list(self._parsers.values())]
        for ix, output in enumerate(outputs):
            output.link(self._internal_sync.inputs[str(ix)])
        self._logger.debug(f"Internal Sync node created with {len(outputs)} outputs")



import inspect
from typing import Callable, List, Tuple

import numpy as np
import pytest

from depthai_nodes.message.utils import copy_message
from tests.utils.messages import creators as message_creators

ATTRS_TO_IGNORE = [
    "Type",  # dai.ImgFrame attribute
]


def equal_attributes(obj1, obj2):
    if type(obj1) is not type(obj2):
        return False
    if np.isscalar(obj1):
        # int, float, str, bool, bytes, np.int32, np.float64, np.bool_, etc.
        return obj1 == obj2
    elif isinstance(obj1, np.ndarray):
        return np.array_equal(obj1, obj2)
    elif isinstance(obj1, (List, Tuple)):
        return len(obj1) == len(obj2) and all(
            equal_attributes(a, b) for a, b in zip(obj1, obj2)
        )
    elif hasattr(obj1, "__dir__"):
        # iterate attributes
        attrs = [
            attr
            for attr in dir(obj1)
            if attr not in ATTRS_TO_IGNORE  # skip ignored attributes
            and not attr.startswith("_")  # skip private attributes
            and not inspect.ismethod(getattr(obj1, attr, None))  # skip methods
        ]
        return all(
            equal_attributes(getattr(obj1, attr, None), getattr(obj2, attr, None))
            for attr in attrs
        )
    else:
        raise ValueError(f"Unsupported attribute type: {type(obj1)}")


@pytest.mark.parametrize(
    "message_creator",
    message_creators.__all__,
)
def test_message_copying(message_creator: Tuple[str, Callable]):
    creator_function = getattr(message_creators, message_creator)

    msg = creator_function()
    try:
        msg_copy = copy_message(msg)
        assert isinstance(msg_copy, type(msg))
        assert equal_attributes(msg, msg_copy)
        if hasattr(msg, "getSequenceNum"):
            assert msg.getSequenceNum() == msg_copy.getSequenceNum()
        if hasattr(msg, "getTimestamp"):
            assert msg.getTimestamp() == msg_copy.getTimestamp()
        if hasattr(msg, "getTimestampDevice"):
            assert msg.getTimestampDevice() == msg_copy.getTimestampDevice()
        if hasattr(msg, "getTransformation"):
            assert equal_attributes(
                msg_copy.getTransformation(), msg.getTransformation()
            )  # comparisson of dai.ImgTransformation objects

    except TypeError:  # copying not implemented for all messages
        pass



import numpy as np

from depthai_nodes.node.parsers.utils.nms import nms


def distance2bbox(points, distance, max_shape=None):
    """Decode distance prediction to bounding box.

    @param points: Shape (n, 2), [x, y].
    @type points: np.ndarray
    @param distance: Distance from the given point to 4 boundaries (left, top, right,
        bottom).
    @type distance: np.ndarray
    @param max_shape: Shape of the image.
    @type max_shape: Tuple[int, int]
    @return: Decoded bboxes.
    @rtype: np.ndarray
    """
    x1 = points[:, 0] - distance[:, 0]
    y1 = points[:, 1] - distance[:, 1]
    x2 = points[:, 0] + distance[:, 2]
    y2 = points[:, 1] + distance[:, 3]
    if max_shape is not None:
        x1 = x1.clamp(min=0, max=max_shape[1])
        y1 = y1.clamp(min=0, max=max_shape[0])
        x2 = x2.clamp(min=0, max=max_shape[1])
        y2 = y2.clamp(min=0, max=max_shape[0])
    return np.stack([x1, y1, x2, y2], axis=-1)


def distance2kps(points, distance, max_shape=None):
    """Decode distance prediction to keypoints.

    @param points: Shape (n, 2), [x, y].
    @type points: np.ndarray
    @param distance: Distance from the given point to 4 boundaries (left, top, right,
        bottom).
    @type distance: np.ndarray
    @param max_shape: Shape of the image.
    @type max_shape: Tuple[int, int]
    @return: Decoded keypoints.
    @rtype: np.ndarray
    """
    preds = []
    for i in range(0, distance.shape[1], 2):
        px = points[:, i % 2] + distance[:, i]
        py = points[:, i % 2 + 1] + distance[:, i + 1]
        if max_shape is not None:
            px = px.clamp(min=0, max=max_shape[1])
            py = py.clamp(min=0, max=max_shape[0])
        preds.append(px)
        preds.append(py)
    return np.stack(preds, axis=-1)


def decode_scrfd(
    bboxes_concatenated,
    scores_concatenated,
    kps_concatenated,
    feat_stride_fpn,
    input_size,
    num_anchors,
    score_threshold,
    nms_threshold,
):
    """Decode the detection results of SCRFD.

    @param bboxes_concatenated: List of bounding box predictions for each scale.
    @type bboxes_concatenated: list[np.ndarray]
    @param scores_concatenated: List of confidence score predictions for each scale.
    @type scores_concatenated: list[np.ndarray]
    @param kps_concatenated: List of keypoint predictions for each scale.
    @type kps_concatenated: list[np.ndarray]
    @param feat_stride_fpn: List of feature strides for each scale.
    @type feat_stride_fpn: list[int]
    @param input_size: Input size of the model.
    @type input_size: tuple[int]
    @param num_anchors: Number of anchors.
    @type num_anchors: int
    @param score_threshold: Confidence score threshold.
    @type score_threshold: float
    @param nms_threshold: Non-maximum suppression threshold.
    @type nms_threshold: float
    @return: Bounding boxes, confidence scores, and keypoints of detected objects.
    @rtype: tuple[np.ndarray, np.ndarray, np.ndarray]
    """
    scores_list = []
    bboxes_list = []
    kps_list = []

    for idx, stride in enumerate(feat_stride_fpn):
        scores = scores_concatenated[idx]
        bbox_preds = bboxes_concatenated[idx] * stride
        kps_preds = kps_concatenated[idx] * stride

        height = input_size[0] // stride
        width = input_size[1] // stride

        anchor_centers = np.stack(np.mgrid[:height, :width][::-1], axis=-1).astype(
            np.float32
        )
        anchor_centers = (anchor_centers * stride).reshape((-1, 2))
        if num_anchors > 1:
            anchor_centers = np.stack([anchor_centers] * num_anchors, axis=1).reshape(
                (-1, 2)
            )

        pos_inds = np.where(scores >= score_threshold)[0]
        bboxes = distance2bbox(anchor_centers, bbox_preds)
        pos_scores = scores[pos_inds]
        pos_bboxes = bboxes[pos_inds]
        scores_list.append(pos_scores.reshape(-1, 1))
        bboxes_list.append(pos_bboxes)

        kpss = distance2kps(anchor_centers, kps_preds)
        kpss = kpss.reshape((kpss.shape[0], -1, 2))
        pos_kpss = kpss[pos_inds]
        kps_list.append(pos_kpss)

    scores = np.vstack(scores_list)
    scores_ravel = scores.ravel()
    order = scores_ravel.argsort()[::-1]
    bboxes = np.vstack(bboxes_list)
    kpss = np.vstack(kps_list)

    pre_det = np.hstack((bboxes, scores)).astype(np.float32, copy=False)
    pre_det = pre_det[order, :]
    keep = nms(pre_det, nms_threshold)
    det = pre_det[keep, :]
    kpss = kpss[order, :, :]
    kpss = kpss[keep, :, :]

    height, width = input_size
    scores = det[:, 4]
    bboxes = det[:, :4] / np.array([width, height] * 2)

    keypoints = kpss / np.tile([width, height], (5, 1))
    keypoints = keypoints.reshape(-1, 5, 2)
    keypoints = np.clip(keypoints, 0, 1)

    return bboxes, scores, keypoints



import depthai as dai
import pytest

from depthai_nodes.node import HostParsingNeuralNetwork, YOLOExtendedParser
from tests.utils import InputMock, NeuralNetworkMock, PipelineMock


@pytest.fixture
def pipeline():
    dai.node.NeuralNetwork = NeuralNetworkMock
    return PipelineMock()


def test_yolo(pipeline: PipelineMock):
    nn_archive = dai.NNArchive(
        dai.getModelFromZoo(dai.NNModelDescription("luxonis/yolov6-nano", "RVC2"))
    )
    nn = pipeline.create(HostParsingNeuralNetwork).build(
        input=InputMock(), nn_source=nn_archive, fps=30
    )
    parser = nn.getParser()
    assert isinstance(parser, YOLOExtendedParser)


def test_unsupported(pipeline: PipelineMock):
    nn_archive = dai.NNArchive(
        dai.getModelFromZoo(
            dai.NNModelDescription("luxonis/mobilenet-ssd:300x300", "RVC2")
        )
    )
    nn = pipeline.create(HostParsingNeuralNetwork)
    with pytest.raises(ValueError):
        nn.build(input=InputMock(), nn_source=nn_archive, fps=30)



from typing import Any, Dict

import depthai as dai

from depthai_nodes.message.creators import create_image_message
from depthai_nodes.node.parsers.base_parser import BaseParser
from depthai_nodes.node.parsers.utils import unnormalize_image


class ImageOutputParser(BaseParser):
    """Parser class for image-to-image models (e.g. DnCNN3, zero-dce etc.) where the
    output is a modifed image (denoised, enhanced etc.).

    Attributes
    ----------
    output_layer_name: str
        Name of the output layer relevant to the parser.
    output_is_bgr : bool
        Flag indicating if the output image is in BGR (Blue-Green-Red) format.

    Output Message/s
    -------
    **Type**: dai.ImgFrame

    **Description**: Image message containing the output image e.g. denoised or enhanced images.

    Error Handling
    --------------
    **ValueError**: If the output is not 3- or 4-dimensional.

    **ValueError**: If the number of output layers is not 1.
    """

    def __init__(
        self, output_layer_name: str = "", output_is_bgr: bool = False
    ) -> None:
        """Initializes the parser node.

        param output_layer_name: Name of the output layer relevant to the parser.
        type output_layer_name: str
        @param output_is_bgr: Flag indicating if the output image is in BGR.
        @type output_is_bgr: bool
        """
        super().__init__()
        self.output_layer_name = output_layer_name
        self.output_is_bgr = output_is_bgr

        self._platform = (
            self.getParentPipeline().getDefaultDevice().getPlatformAsString()
        )
        self._logger.debug(
            f"ImageOutputParser initialized with output_layer_name='{output_layer_name}', output_is_bgr={output_is_bgr}"
        )

    def setOutputLayerName(self, output_layer_name: str) -> None:
        """Sets the name of the output layer.

        @param output_layer_name: The name of the output layer.
        @type output_layer_name: str
        """
        if not isinstance(output_layer_name, str):
            raise ValueError("Output layer name must be a string.")
        self.output_layer_name = output_layer_name
        self._logger.debug(f"Output layer name set to '{self.output_layer_name}'")

    def setBGROutput(self) -> None:
        """Sets the flag indicating that output image is in BGR."""
        self.output_is_bgr = True
        self._logger.debug(f"Output is BGR set to {self.output_is_bgr}")

    def build(
        self,
        head_config: Dict[str, Any],
    ) -> "ImageOutputParser":
        """Configures the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: ImageOutputParser
        """

        output_layers = head_config.get("outputs", [])
        if len(output_layers) != 1:
            raise ValueError(
                f"MapOutputParser expects exactly 1 output layers, got {output_layers} layers."
            )
        self.output_layer_name = output_layers[0]
        self.output_is_bgr = head_config.get("output_is_bgr", self.output_is_bgr)

        self._logger.debug(
            f"ImageOutputParser built with output_layer_name='{self.output_layer_name}', output_is_bgr={self.output_is_bgr}"
        )

        return self

    def run(self):
        self._logger.debug("ImageOutputParser run started")
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            layers = output.getAllLayerNames()
            self._logger.debug(f"Processing input with layers: {layers}")
            if len(layers) == 1 and self.output_layer_name == "":
                self.output_layer_name = layers[0]
            elif len(layers) != 1 and self.output_layer_name == "":
                raise ValueError(
                    f"Expected 1 output layer, got {len(layers)} layers. Please provide the output_layer_name."
                )

            output_image = output.getTensor(self.output_layer_name, dequantize=True)

            if output_image.shape[0] == 1:
                output_image = output_image[0]  # remove batch dimension

            if len(output_image.shape) != 3:
                raise ValueError(
                    f"Expected 3D output tensor, got {len(output_image.shape)}D."
                )

            image = unnormalize_image(output_image)

            image_message = create_image_message(
                image=image,
                is_bgr=self.output_is_bgr,
                img_frame_type=(
                    dai.ImgFrame.Type.BGR888p
                    if self._platform == "RVC2"
                    else dai.ImgFrame.Type.BGR888i
                ),
            )
            image_message.setTimestamp(output.getTimestamp())
            if output.getTransformation():
                image_message.setTransformation(output.getTransformation())
            image_message.setSequenceNum(output.getSequenceNum())

            self._logger.debug(f"Created image message with shape {image.shape}")

            self.out.send(image_message)

            self._logger.debug("Image message sent successfully")
